/*
* TGPR: Transfer Learning Based Visual Tracking with Gaussian Process Regression
*
* Code to accompany the paper:
*   Transfer Learning Based Visual Tracking with Gaussian Process Regression
*   Jin Gao, Haibin Ling, Weiming Hu, Junliang Xing
*   ECCV2014, Part III, LNC 8691, pp. 188-203, 2014
*
* Copyright (C) 2014 Jin Gao
*
* This file is part of TGPR.
*
*
*The whole framework of this code is based on the framework of the publicly available code:
*
* Struck: Structured Output Tracking with Kernels
*
* Struck is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* Struck is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with Struck.  If not, see <http://www.gnu.org/licenses/>.
*
*/

#include "CovFeatures.h"
#include "Config.h"
#include "Sample.h"
#include "Rect.h"
#define EIGEN_NO_DEBUG
#include <iostream>
#include <Eigen/Eigenvalues> 
#include <cmath>

using namespace Eigen;
using namespace cv;
using namespace std;

static const int numCovFeature = 6;
static const int kNumBins = (numCovFeature*numCovFeature + numCovFeature) / 2.;
static const int kNumLevels = 1;

CovFeatures::CovFeatures(const Config& conf) :
kNumCellsX(conf.Nblock),
kNumCellsY(conf.Nblock)
{
	int nc = 0;
	for (int i = 0; i < kNumLevels; ++i)
	{
		nc += (i + kNumCellsX)*(i + kNumCellsY);
	}
	SetCount(kNumBins*nc);
}

void CovFeatures::GetkNumBinsFeatures()
{
	kNumBinsFeatures = kNumBins;
}

void CovFeatures::UpdateFeatureVector(const Sample& s, Eigen::VectorXd& featVect)
{
	IntRect rect = s.GetROI(); // note this truncates to integers
	
	m_featVec.setZero();
	featVect = VectorXd::Zero(m_featVec.size());
	MatrixXd logCov = MatrixXd::Zero(numCovFeature, numCovFeature);
	//int histind = 0;

	for (int il = 0; il < kNumLevels; ++il)
	{
		int ncx = il + kNumCellsX;
		int ncy = il + kNumCellsY;
		float w = s.GetROI().Width() / ncx;
		float h = s.GetROI().Height() / ncy;
		FloatRect cell(0.f, 0.f, w, h);
		int ix, iy;
		//Eigen::VectorXd m_featVec_tmp = VectorXd::Zero(m_featVec.size());
		//const Sample* p_s(&s);
		//Sample s_tmp(s.GetImage(), s.GetROI());
		//m_featVec_tmp.setZero();
		//#pragma omp parallel for private(ix,iy) firstprivate(cell,logCov)
		for (int histind = 0; histind < ncx*ncy; ++histind)
		//for (int iy = 0; iy < ncy; ++iy)
		{
			iy = histind % ncy;
			ix = (int)((histind - iy) / ncx);
			cell.SetYMin(s.GetROI().YMin() + iy*h);
			//for (int ix = 0; ix < ncx; ++ix)
			//{
			cell.SetXMin(s.GetROI().XMin() + ix*w);
			logm(s.GetImage().Cov(cell), logCov);
#if VERBOSEMY2		
				cout << "Here is the matrix logCov:" << endl << logCov << endl;
#endif
				int step = 0;
				for (int i = 0; i < numCovFeature; ++i)
				{
					featVect.segment(histind*kNumBins + step, numCovFeature - i) = logCov.block(i, i, numCovFeature - i, 1);
					step += (numCovFeature - i);
				}
				//++histind;
			//}
		}
		//m_featVec = m_featVec_tmp;
	}
}

void CovFeatures::logm(const cv::Mat& Cov, Eigen::MatrixXd& logCov) //log of matrix
{
	//#pragma omp critical
	//{
	MatrixXd A = MatrixXd::Zero(numCovFeature, numCovFeature);
	MatrixXd logD = MatrixXd::Zero(numCovFeature, numCovFeature);
	for (int i = 0; i < numCovFeature; ++i)
	{
		for (int j = 0; j < numCovFeature; ++j)
		{
			A(i, j) = Cov.at<double>(i, j);
			if (i == j)
			{
				A(i, j) += 1e-5;
			}
		}
	}
	EigenSolver<MatrixXd> es(A);
	MatrixXd D = es.pseudoEigenvalueMatrix();
	MatrixXd V = es.pseudoEigenvectors();
#if VERBOSEMY		
	cout << "Here is the matrix D:" << endl << D << endl;
#endif
	for (int ind = 0; ind < numCovFeature; ++ind)
	{
		if (D(ind, ind) < 1e-9)
		{
			D(ind, ind) = -D(ind, ind) + 1e-9;
			cout << "1" << endl;
		}
		logD(ind, ind) = log(D(ind, ind));
	}
	logCov = V*logD*V.inverse();

#if VERBOSEMY2		
	cout << "Here is the matrix logCov:" << endl << logCov << endl;
	cout << "Here is the matrix D:" << endl << D << endl;
	cout << "Here is the matrix logD:" << endl << logD << endl;
	cout << "Here is the matrix V:" << endl << V << endl;
	cout << "Here is the matrix V.inverse():" << endl << V.inverse() << endl;
	cout << "Here is the matrix A:" << endl << A << endl;
#endif
	//}
}