/*
* TGPR: Transfer Learning Based Visual Tracking with Gaussian Process Regression
*
* Code to accompany the paper:
*   Transfer Learning Based Visual Tracking with Gaussian Process Regression
*   Jin Gao, Haibin Ling, Weiming Hu, Junliang Xing
*   ECCV2014, Part III, LNC 8691, pp. 188-203, 2014
*
* Copyright (C) 2014 Jin Gao
*
* This file is part of TGPR.
*
*
*The whole framework of this code is based on the framework of the publicly available code:
*
* Struck: Structured Output Tracking with Kernels
*
* Struck is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* Struck is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with Struck.  If not, see <http://www.gnu.org/licenses/>.
*
*/

#include "GPR.h"

#include "Config.h"
#include "Features.h"
#include "Kernels.h"
#include "Sample.h"
#include "Rect.h"
#include "Timing.h"
#include "GraphUtils/GraphUtils.h"
#include <opencv/cv.h>
#define EIGEN_NO_DEBUG
#include <Eigen/Dense> 
#include <Eigen/Core>


#include <opencv/highgui.h>

#include <sys/timeb.h>
#include <ctime>
#include <climits>

static const int kTileSizeGPR = 30;
static const int n_Tposi = 20;
static const int n_Aposi = 54;
static const int n_Anega = 54 * 4;
using namespace cv;

using namespace std;
using namespace Eigen;

static const int kMaxSVsGPR = 2000; // TODO (only used when no budget)


GPR::GPR(const Config& conf, const Features& features, const Kernel& kernel, Timing& Timer) :
m_config(conf),
m_pTiming(&Timer),
m_features(features),
m_kernel(kernel)
{
	int N = 1000;
	m_K = MatrixXd::Zero(N, N);
	m_debugImage = Mat(800, 600, CV_8UC3);
	m_debugImage1 = Mat(800, 600, CV_8UC3);
	m_debugImage2 = Mat(800, 600, CV_8UC3);
	flag = 0;
	kNumBins = 0;
}

GPR::~GPR()
{
}

Eigen::MatrixXd GPR::ComputInvers(const Eigen::MatrixXd M)
{
	MatrixXd inversM;
	//Eigen::FullPivLU<MatrixXd> lu(M);
	//if (lu.isInvertible())
	//{
	//	inversM = lu.inverse();
	//}
	inversM = M.inverse();
	return inversM;
}

Eigen::MatrixXd GPR::SetZerosInW(const Eigen::MatrixXd W, int knn)
{
	Eigen::MatrixXd Wr;
	Eigen::MatrixXd WrForSort;
	Wr = W;
	WrForSort = W.transpose();
	double* pWr;
#if VERBOSEMY2		
	cout << "Here is the matrix W:" << endl << W << endl;
#endif
	pWr = (double*)(WrForSort.data());
	//Mat Wcv = Mat(W.rows(), W.cols(), CV_32F);
	//for (int i = 0; i < W.rows(); ++i)
	//{
	//	for (int j = 0; j < W.cols(); ++j)
	//	{
	//		Wcv.at<float>(i, j) = W(i, j);
	//	}
	//}
	//Mat sorted, indexkernel;

	//sortIdx(Wcv, indexkernel, CV_SORT_EVERY_ROW + CV_SORT_DESCENDING);
	//cv::sort(Wcv, sorted, CV_SORT_EVERY_ROW + CV_SORT_DESCENDING);
	int n = WrForSort.rows();
	VectorXd kNNWr = VectorXd::Zero(n);
	for (int i = 0; i < W.rows(); ++i)
	{
		kNNWr(i) = GetMinKth(pWr, i*n, (i + 1)*n - 1, n-knn);
		for (int j = 0; j < W.cols(); ++j)
		{
			if ((W(i, j) - kNNWr(i)) < 0)
			{
				Wr(i, j) = 0.0;
			}
		}
	}
#if VERBOSEMY2		
	cout << "Here is the matrix Wr:" << endl << Wr << endl;
#endif
	return Wr;
}

void GPR::Eval(const MultiSample& sample, std::vector<int>& FIVE, bool& flagcase)
{
	SupportPattern* sp = new SupportPattern;
	const vector<FloatRect>& rects = sample.GetRects();
	FloatRect centre = rects[0];
	// evaluate features for each sample
	sp->x.resize(rects.size());

	ftime(&(m_pTiming->startTimer6));
	const_cast<Features&>(m_features).Eval(sample, sp->x);
	ftime(&(m_pTiming->endTimer6));
	m_pTiming->storetime(m_pTiming->TimeStore6, m_pTiming->startTimer6, m_pTiming->endTimer6);

	Unlabeled_sps->x.erase(Unlabeled_sps->x.begin(), Unlabeled_sps->x.end());
	Unlabeled_sps->yv.erase(Unlabeled_sps->yv.begin(), Unlabeled_sps->yv.end());
	Unlabeled_sps->images.erase(Unlabeled_sps->images.begin(), Unlabeled_sps->images.end());

	for (int i = 0; i < (int)rects.size(); ++i)
	{
		// express r in coord frame of centre sample
		FloatRect r = rects[i];
		r.Translate(-centre.XMin(), -centre.YMin());
		Unlabeled_sps->yv.push_back(r);
		if (!m_config.quietMode && m_config.debugMode)
		{
			// store a thumbnail for each sample
			Mat im(kTileSizeGPR, kTileSizeGPR, CV_8UC1);
			IntRect rect = rects[i];
			cv::Rect roi(rect.XMin(), rect.YMin(), rect.Width(), rect.Height());
			cv::resize(sample.GetImage().GetImage(0)(roi), im, im.size());
			Unlabeled_sps->images.push_back(im);
		}
		Unlabeled_sps->x.push_back(sp->x[i]);
	}

	Unlabeled_sps->y = 0;
	Unlabeled_sps->refCount = 0;


	costruSampleSet(); //construct the sample set for GPR

	ftime(&(m_pTiming->startTimer3));
	//Eigen::setNbThreads(4);
	costruKernel(); //calculate W
	//Eigen::setNbThreads(1);
	ftime(&(m_pTiming->endTimer3));
	m_pTiming->storetime(m_pTiming->TimeStore3, m_pTiming->startTimer3, m_pTiming->endTimer3);

	int kersize((int)SampleSet_sps->x.size());
	MatrixXd W(m_K.block(0, 0, kersize, kersize));
	int knn = 50;

	ftime(&(m_pTiming->startTimer4));
	W = SetZerosInW(W, knn);
	ftime(&(m_pTiming->endTimer4));
	m_pTiming->storetime(m_pTiming->TimeStore4, m_pTiming->startTimer4, m_pTiming->endTimer4);
	//for (int i = 0; i < kersize; ++i)
	//{
	//	W.row(i) = W.row(i)/W.row(i).sum();
	//}
#if VERBOSEMY3		
	cout << "Here is the matrix W.block:" << endl << W.block(0, 0, kersize, 5) << endl;
#endif
	float beta = 1.0;
	float sig = 1000.0;
	float r = 10.0;
	float eta = 0.4;
	int pool = 40;

	//ftime(&(m_pTiming->startTimer5));
	if (flag == 0) //flag==0, only use Target Classifier for tracking
	{
		MatrixXd D = MatrixXd::Zero(kersize, kersize);
		MatrixXd eye = MatrixXd::Zero(kersize, kersize);
//#pragma omp parallel for firstprivate(D, W, eye)
		for (int i = 0; i < kersize; ++i)
		{
			D(i, i) = W.row(i).sum();
			eye(i, i) = 1;
		}
		MatrixXd delta = D - W;
		MatrixXd Mdelta = beta*(delta + eye*(1 / (sig*sig)));
		MatrixXd G = ComputInvers(Mdelta);
		m_K.block(0, 0, kersize, kersize) = W;
		int tP = (int)Tp_sps->x.size();
		int tN = (int)Tn_sps->x.size();
		int un = (int)Unlabeled_sps->x.size();
		MatrixXd invGLL = ComputInvers(G.block(0, 0, tP + tN, tP + tN));
		VectorXd Z = VectorXd::Ones(tP + tN);
		VectorXd One = Z;
		Z.tail(tN) *= (-1);
		VectorXd YL = Z;
		MatrixXd Q = MatrixXd::Zero(tP + tN, tP + tN);
		VectorXd pai;
		VectorXd dq1q2;
		VectorXd temp;
		MatrixXd H,H_tmp;
		for (int iteration = 0; iteration < 15; ++iteration)
		{
			pai = (One + (-2 * r*YL).array().exp().matrix()).cwiseInverse();
			dq1q2 = r*(Z - One) + 2 * r*(One - pai) - invGLL*YL;
			temp = 4 * r*r*(pai.cwiseProduct(One - pai));
			for (int i = 0; i < (tP + tN); ++i)
			{
				Q(i, i) = temp(i);
			}
			H = -invGLL - Q;
			H_tmp = H.inverse();
			H = H_tmp;
			//H = ComputInvers(H);
			YL = YL - eta*(H*dq1q2);
		}
#if VERBOSEMY4		
		cout << "Here is the matrix YL:" << endl << YL << endl;
#endif
		VectorXd YU = G.block(tP + tN, 0, un, tP + tN)*(invGLL*YL);
		//Mat results = Mat(un, 1, CV_32F);
		Mat prob;
		double *presults;
		presults = (double *)YU.data();
		Mat results = Mat(un, 1, CV_64F, presults);
		sortIdx(results, Index, CV_SORT_EVERY_COLUMN + CV_SORT_DESCENDING);
		cv::sort(results, prob, CV_SORT_EVERY_COLUMN + CV_SORT_DESCENDING);
		Index1 = Index;
		for (int i = 0; i < 5; ++i)
		{
			FIVE.push_back(Index.at<int>(i, 0));
		}
	}
	else //flag == 1, use Auxiliary Classifier and Target Classifier
	{
		ftime(&(m_pTiming->startTimer5));
		int repeat = (int)(n_Tposi / m_config.interval) + 1;
		MatrixXd D = MatrixXd::Zero(kersize, kersize);
		MatrixXd eye = MatrixXd::Zero(kersize, kersize);
//#pragma omp parallel for firstprivate(D, W, eye)
		for (int i = 0; i < kersize; ++i)
		{
			D(i, i) = W.row(i).sum();
			eye(i, i) = 1;
		}
		MatrixXd delta = D - W;
		MatrixXd Mdelta = beta*(delta + eye*(1 / (sig*sig)));
		//MatrixXd G = ComputInvers(Mdelta);  //Alog. 1 line 9
		MatrixXd G = Mdelta.inverse();  //Alog. 1 line 9
		m_K.block(0, 0, kersize, kersize) = W;
		int tP = (int)Tp_sps->x.size();
		int tN = (int)Tn_sps->x.size();
		int aP = (int)Ap_sps->x.size() - repeat;
		int aN = (int)An_sps->x.size() - repeat * 4;
		int un = (int)Unlabeled_sps->x.size();
		//MatrixXd invGLL = ComputInvers(G.block(0, 0, aP + aN + tP + tN, aP + aN + tP + tN));
		MatrixXd invGLL = G.block(0, 0, aP + aN + tP + tN, aP + aN + tP + tN).inverse();
		MatrixXd H11 = invGLL.block(0, 0, aP + aN, aP + aN);
		MatrixXd H12 = invGLL.block(0, aP + aN, aP + aN, tP + tN);
		MatrixXd H21 = invGLL.block(aP + aN, 0, tP + tN, aP + aN);
		MatrixXd H22 = invGLL.block(aP + aN, aP + aN, tP + tN, tP + tN);

		VectorXd Z = VectorXd::Ones(aP + aN + tP + tN);
		VectorXd One = VectorXd::Ones(aP + aN);
		Z.tail(tN) *= (-1);
		Z.segment(aP, aN) *= (-1);
		VectorXd Ya = Z.head(aP + aN);
		VectorXd Ya_initial = Ya;
		VectorXd Yt = Z.tail(tP + tN);
		MatrixXd Q = MatrixXd::Zero(aP + aN, aP + aN);
		MatrixXd H, H_tmp;
		for (int iteration = 0; iteration < 15; ++iteration)
		{
			VectorXd pai = (One + (-2 * r*Ya).array().exp().matrix()).cwiseInverse();
			VectorXd dq1q2 = r*(Ya_initial - One) + 2 * r*(One - pai) - H11*Ya - 0.5*H21.transpose()*Yt - 0.5*H12*Yt;
			VectorXd temp = 4 * r*r*(pai.cwiseProduct(One - pai));
			for (int i = 0; i < (aP + aN); ++i)
			{
				Q(i, i) = temp(i);
			}
			H = -H11 - Q;
			H_tmp = H.inverse();
			H = H_tmp;
			//H = ComputInvers(H);
			Ya = Ya - eta*(H*dq1q2);
		}                                                  //Alog. 1 line 10
#if VERBOSEMY4		
		cout << "Here is the matrix Ya:" << endl << Ya << endl;
		//cout << "Here is the block:" << endl << W.block(0, 0, aP + aN, aP + aN) << endl;
		//cout << "Here is the block:" << endl << G.block(0, 0, aP + aN + tP + tN, aP + aN + tP + tN) << endl;
		//cout << "Here is the block:" << endl << H11 << endl;
		//cout << "Here is the block:" << endl << H22 << endl;
#endif

		/*MatrixXd Wa(MatrixXd::Zero(aP + aN + un, aP + aN + un));
		Wa.block(0, 0, aP + aN, aP + aN) = W.block(0, 0, aP + aN, aP + aN);
		Wa.block(0, aP + aN, aP + aN, un) = W.block(0, aP + aN + tP + tN, aP + aN, un);
		Wa.block(aP + aN, 0, un, aP + aN) = W.block(aP + aN + tP + tN, 0, un, aP + aN);
		Wa.block(aP + aN, aP + aN, un, un) = W.block(aP + aN + tP + tN, aP + aN + tP + tN, un, un); //Alog. 1 line 11
		MatrixXd Da = MatrixXd::Zero(aP + aN + un, aP + aN + un);
		MatrixXd eyea = MatrixXd::Zero(aP + aN + un, aP + aN + un);
		for (int i = 0; i < (aP + aN + un); ++i)
		{
			Da(i, i) = Wa.row(i).sum();
			eyea(i, i) = 1;
		}
		MatrixXd deltaa = Da - Wa;
		MatrixXd Mdeltaa = beta*(deltaa + eyea*(1 / (sig*sig)));
		MatrixXd Ga = Mdeltaa.inverse();
		MatrixXd invGLLa = Ga.block(0, 0, aP + aN, aP + aN).inverse();
		//MatrixXd Ga = ComputInvers(Mdeltaa);
		//MatrixXd invGLLa = ComputInvers(Ga.block(0, 0, aP + aN, aP + aN));
		VectorXd YUa = Ga.block(aP + aN, 0, un, aP + aN)*(invGLLa*Ya); //Alog. 1 line 12

		MatrixXd Wt(W.block(aP + aN, aP + aN, tP + tN + un, tP + tN + un)); //Alog. 1 line 14
		MatrixXd Dt = MatrixXd::Zero(tP + tN + un, tP + tN + un);
		MatrixXd eyet = MatrixXd::Zero(tP + tN + un, tP + tN + un);
		for (int i = 0; i < (tP + tN + un); ++i)
		{
			Dt(i, i) = Wt.row(i).sum();
			eyet(i, i) = 1;
		}
		MatrixXd deltat = Dt - Wt;
		MatrixXd Mdeltat = beta*(deltat + eyet*(1 / (sig*sig)));
		MatrixXd Gt = Mdeltat.inverse();
		MatrixXd invGLLt = Gt.block(0, 0, tP + tN, tP + tN).inverse();
		//MatrixXd Gt = ComputInvers(Mdeltat);
		//MatrixXd invGLLt = ComputInvers(Gt.block(0, 0, tP + tN, tP + tN));
		VectorXd YUt = Gt.block(tP + tN, 0, un, tP + tN)*(invGLLt*Yt); //Alog. 1 line 16

		ftime(&(m_pTiming->endTimer5));
		m_pTiming->storetime(m_pTiming->TimeStore5, m_pTiming->startTimer5, m_pTiming->endTimer5);

		//Mat resultsA = Mat(un, 1, CV_32F);
		Mat IndexA;
		Mat probA;
		//Mat resultsT = Mat(un, 1, CV_32F);
		Mat IndexT;
		Mat probT;
		double *presultsA;
		double *presultsT;
		presultsA = (double *)YUa.data();
		presultsT = (double *)YUt.data();
		Mat resultsA = Mat(un, 1, CV_64F, presultsA);
		Mat resultsT = Mat(un, 1, CV_64F, presultsT);
		sortIdx(resultsA, IndexA, CV_SORT_EVERY_COLUMN + CV_SORT_DESCENDING);
		sortIdx(resultsT, IndexT, CV_SORT_EVERY_COLUMN + CV_SORT_DESCENDING);
		cv::sort(resultsA, probA, CV_SORT_EVERY_COLUMN + CV_SORT_DESCENDING);
		cv::sort(resultsT, probT, CV_SORT_EVERY_COLUMN + CV_SORT_DESCENDING);*/

		Mat IndexA;
		Mat probA;
		Mat IndexT;
		Mat probT;
		Mat resultsA;
		Mat resultsT;

		MatrixXd Wa(MatrixXd::Zero(aP + aN + un, aP + aN + un));
		MatrixXd Da = MatrixXd::Zero(aP + aN + un, aP + aN + un);
		MatrixXd eyea = MatrixXd::Zero(aP + aN + un, aP + aN + un);
		MatrixXd deltaa;
		MatrixXd Mdeltaa;
		MatrixXd Ga;
		MatrixXd invGLLa;
		VectorXd YUa;
		double *presultsA;
		int ia;

		MatrixXd Wt; //Alog. 1 line 14
		MatrixXd Dt = MatrixXd::Zero(tP + tN + un, tP + tN + un);
		MatrixXd eyet = MatrixXd::Zero(tP + tN + un, tP + tN + un);
		MatrixXd deltat;
		MatrixXd Mdeltat;
		MatrixXd Gt;
		MatrixXd invGLLt;
		VectorXd YUt;
		double *presultsT;
		int it;

#pragma omp parallel for firstprivate(aP, aN, un, W, beta, sig)
		for (int index_classifier = 0; index_classifier < 2; ++index_classifier)
		{
			if (index_classifier == 0)
			{
				Wa.block(0, 0, aP + aN, aP + aN) = W.block(0, 0, aP + aN, aP + aN);
				Wa.block(0, aP + aN, aP + aN, un) = W.block(0, aP + aN + tP + tN, aP + aN, un);
				Wa.block(aP + aN, 0, un, aP + aN) = W.block(aP + aN + tP + tN, 0, un, aP + aN);
				Wa.block(aP + aN, aP + aN, un, un) = W.block(aP + aN + tP + tN, aP + aN + tP + tN, un, un); //Alog. 1 line 11
				for (ia = 0; ia < (aP + aN + un); ++ia)
				{
					Da(ia, ia) = Wa.row(ia).sum();
					eyea(ia, ia) = 1;
				}
				deltaa = Da - Wa;
				Mdeltaa = beta*(deltaa + eyea*(1 / (sig*sig)));
				Ga = Mdeltaa.inverse();
				invGLLa = Ga.block(0, 0, aP + aN, aP + aN).inverse();
				//MatrixXd Ga = ComputInvers(Mdeltaa);
				//MatrixXd invGLLa = ComputInvers(Ga.block(0, 0, aP + aN, aP + aN));
				YUa = Ga.block(aP + aN, 0, un, aP + aN)*(invGLLa*Ya); //Alog. 1 line 12

				presultsA = (double *)YUa.data();
				resultsA = Mat(un, 1, CV_64F, presultsA);
				sortIdx(resultsA, IndexA, CV_SORT_EVERY_COLUMN + CV_SORT_DESCENDING);
				cv::sort(resultsA, probA, CV_SORT_EVERY_COLUMN + CV_SORT_DESCENDING);
			}
			else if (index_classifier == 1)
			{
				Wt = W.block(aP + aN, aP + aN, tP + tN + un, tP + tN + un);
				for (it = 0; it < (tP + tN + un); ++it)
				{
					Dt(it, it) = Wt.row(it).sum();
					eyet(it, it) = 1;
				}
				deltat = Dt - Wt;
				Mdeltat = beta*(deltat + eyet*(1 / (sig*sig)));
				Gt = Mdeltat.inverse();
				invGLLt = Gt.block(0, 0, tP + tN, tP + tN).inverse();
				//MatrixXd Gt = ComputInvers(Mdeltat);
				//MatrixXd invGLLt = ComputInvers(Gt.block(0, 0, tP + tN, tP + tN));
				YUt = Gt.block(tP + tN, 0, un, tP + tN)*(invGLLt*Yt); //Alog. 1 line 16

				presultsT = (double *)YUt.data();
				resultsT = Mat(un, 1, CV_64F, presultsT);
				sortIdx(resultsT, IndexT, CV_SORT_EVERY_COLUMN + CV_SORT_DESCENDING);
				cv::sort(resultsT, probT, CV_SORT_EVERY_COLUMN + CV_SORT_DESCENDING);
			}
		}
		ftime(&(m_pTiming->endTimer5));
		m_pTiming->storetime(m_pTiming->TimeStore5, m_pTiming->startTimer5, m_pTiming->endTimer5);

		/*//Mat resultsA = Mat(un, 1, CV_32F);
		Mat IndexA;
		Mat probA;
		//Mat resultsT = Mat(un, 1, CV_32F);
		Mat IndexT;
		Mat probT;
		double *presultsA;
		double *presultsT;
		presultsA = (double *)YUa.data();
		presultsT = (double *)YUt.data();
		Mat resultsA = Mat(un, 1, CV_64F, presultsA);
		Mat resultsT = Mat(un, 1, CV_64F, presultsT);
		sortIdx(resultsA, IndexA, CV_SORT_EVERY_COLUMN + CV_SORT_DESCENDING);
		sortIdx(resultsT, IndexT, CV_SORT_EVERY_COLUMN + CV_SORT_DESCENDING);
		cv::sort(resultsA, probA, CV_SORT_EVERY_COLUMN + CV_SORT_DESCENDING);
		cv::sort(resultsT, probT, CV_SORT_EVERY_COLUMN + CV_SORT_DESCENDING);*/

		Index = IndexA;
		Index1 = IndexT;

		vector<int> Fua;
		vector<int> Fut;
		int numI = 0;

		for (int i = 0; i < pool; ++i)
		{
			if (resultsT.at<double>(IndexA.at<int>(i, 0), 0) - probT.at<double>(pool, 0) > 1e-8)
			{
				numI++;
				Fua.push_back(IndexA.at<int>(i, 0));
			}
			if (resultsA.at<double>(IndexT.at<int>(i, 0), 0) - probA.at<double>(pool, 0) > 1e-8)
			{
				Fut.push_back(IndexT.at<int>(i, 0));
			}
		}

		if (numI > (pool / 2))
		{
			FIVE.insert(FIVE.begin(), Fua.begin(), Fua.begin() + 5);
			flagcase = false;
		}
		else
		{
			flagcase = true;
			if (Fut.size() == 0)
			{
				for (int i = 0; i < 5; ++i)
				{
					FIVE.push_back(IndexA.at<int>(i, 0));
				}
			}
			else
			{
				FIVE = Fut;
			}
		}
	}
	//ftime(&(m_pTiming->endTimer5));
	//m_pTiming->storetime(m_pTiming->TimeStore5, m_pTiming->startTimer5, m_pTiming->endTimer5);
}

void GPR::UpdateT(const MultiSample& sample, int y)
{
	SupportPattern* sp = new SupportPattern;
	const vector<FloatRect>& rects = sample.GetRects();
	FloatRect centre = rects[y];
	// evaluate features for each sample
	sp->x.resize(rects.size());
	const_cast<Features&>(m_features).Eval(sample, sp->x);
	Tn_sps->x.erase(Tn_sps->x.begin(), Tn_sps->x.end());
	Tn_sps->yv.erase(Tn_sps->yv.begin(), Tn_sps->yv.end());
	Tn_sps->images.erase(Tn_sps->images.begin(), Tn_sps->images.end());
	for (int i = 0; i < (int)rects.size(); ++i)
	{
		if (i == 0)
		{
			// express r in coord frame of centre sample
			FloatRect r = rects[i];
			r.Translate(-centre.XMin(), -centre.YMin());
			Tp_sps->yv.push_back(r);
			if (!m_config.quietMode && m_config.debugMode)
			{
				// store a thumbnail for each sample
				Mat im(kTileSizeGPR, kTileSizeGPR, CV_8UC1);
				IntRect rect = rects[i];
				cv::Rect roi(rect.XMin(), rect.YMin(), rect.Width(), rect.Height());
				cv::resize(sample.GetImage().GetImage(0)(roi), im, im.size());
				Tp_sps->images.push_back(im);
			}
			Tp_sps->x.push_back(sp->x[i]);
		}
		else
		{
			// express r in coord frame of centre sample
			FloatRect r = rects[i];
			r.Translate(-centre.XMin(), -centre.YMin());
			Tn_sps->yv.push_back(r);
			if (!m_config.quietMode && m_config.debugMode)
			{
				// store a thumbnail for each sample
				Mat im(kTileSizeGPR, kTileSizeGPR, CV_8UC1);
				IntRect rect = rects[i];
				cv::Rect roi(rect.XMin(), rect.YMin(), rect.Width(), rect.Height());
				cv::resize(sample.GetImage().GetImage(0)(roi), im, im.size());
				Tn_sps->images.push_back(im);
			}
			Tn_sps->x.push_back(sp->x[i]);
		}
	}

	Tp_sps->y = y;
	Tp_sps->refCount = 0;
	Tn_sps->y = y;
	Tn_sps->refCount = 0;

	BudgetMaintenanceGPR();
}

void GPR::UpdateA(const MultiSample& sample, int y)
{
	SupportPattern* sp = new SupportPattern;
	const vector<FloatRect>& rects = sample.GetRects();
	FloatRect centre = rects[y];
	// evaluate features for each sample
	sp->x.resize(rects.size());
	const_cast<Features&>(m_features).Eval(sample, sp->x);
	for (int i = 0; i < (int)rects.size(); ++i)
	{
		if (i == 0)
		{
			// express r in coord frame of centre sample
			FloatRect r = rects[i];
			r.Translate(-centre.XMin(), -centre.YMin());
			Ap_sps->yv.push_back(r);
			if (!m_config.quietMode && m_config.debugMode)
			{
				// store a thumbnail for each sample
				Mat im(kTileSizeGPR, kTileSizeGPR, CV_8UC1);
				IntRect rect = rects[i];
				cv::Rect roi(rect.XMin(), rect.YMin(), rect.Width(), rect.Height());
				cv::resize(sample.GetImage().GetImage(0)(roi), im, im.size());
				Ap_sps->images.push_back(im);
			}
			Ap_sps->x.push_back(sp->x[i]);
		}
		else
		{
			// express r in coord frame of centre sample
			FloatRect r = rects[i];
			r.Translate(-centre.XMin(), -centre.YMin());
			An_sps->yv.push_back(r);
			if (!m_config.quietMode && m_config.debugMode)
			{
				// store a thumbnail for each sample
				Mat im(kTileSizeGPR, kTileSizeGPR, CV_8UC1);
				IntRect rect = rects[i];
				cv::Rect roi(rect.XMin(), rect.YMin(), rect.Width(), rect.Height());
				cv::resize(sample.GetImage().GetImage(0)(roi), im, im.size());
				An_sps->images.push_back(im);
			}
			An_sps->x.push_back(sp->x[i]);
		}
	}

	Ap_sps->y = y;
	Ap_sps->refCount = 0;
	An_sps->y = y;
	An_sps->refCount = 0;

	BudgetMaintenanceGPR();
}

void GPR::BudgetMaintenanceGPR()
{
	while (((int)Tp_sps->x.size() > n_Tposi && flag == 1) || (int)Ap_sps->x.size() > n_Aposi || (int)An_sps->x.size() > n_Anega)
	{
		//cout << "4" << endl;
		BudgetMaintenanceRemoveGPR();
	}
}

void GPR::BudgetMaintenanceRemoveGPR()
{
	if ((int)Tp_sps->x.size() > n_Tposi && flag == 1) //Target Positive samples Buffer size 10
	{
		Tp_sps->x.erase(Tp_sps->x.begin(), Tp_sps->x.end() - n_Tposi);
		Tp_sps->yv.erase(Tp_sps->yv.begin(), Tp_sps->yv.end() - n_Tposi);
		if (!m_config.quietMode && m_config.debugMode)
		{
			Tp_sps->images.erase(Tp_sps->images.begin(), Tp_sps->images.end() - n_Tposi);
		}
	}

	if ((int)Ap_sps->x.size() > n_Aposi) //Auxiliary Positive samples Buffer size 52
	{
		Ap_sps->x.erase(Ap_sps->x.begin(), Ap_sps->x.end() - n_Aposi);
		Ap_sps->yv.erase(Ap_sps->yv.begin(), Ap_sps->yv.end() - n_Aposi);
		if (!m_config.quietMode && m_config.debugMode)
		{
			Ap_sps->images.erase(Ap_sps->images.begin(), Ap_sps->images.end() - n_Aposi);
		}
	}

	if ((int)An_sps->x.size() > n_Anega) //Auxiliary Negative samples Buffer size 52*4
	{
		An_sps->x.erase(An_sps->x.begin(), An_sps->x.end() - n_Anega);
		An_sps->yv.erase(An_sps->yv.begin(), An_sps->yv.end() - n_Anega);
		if (!m_config.quietMode && m_config.debugMode)
		{
			An_sps->images.erase(An_sps->images.begin(), An_sps->images.end() - n_Anega);
		}
	}

}

void GPR::Debug()
{
	cout << Tp_sps->x.size() << "/" << Tn_sps->x.size() << " support patterns/vectors" << endl;
	UpdateDebugImage();
	UpdateDebugImage1();
	UpdateDebugImage2();
	imshow("sampler", m_debugImage);
	imshow("sampler1", m_debugImage1);
	imshow("sampler2", m_debugImage2);
}

void GPR::costruSampleSet()
{
	if (((int)Ap_sps->x.size() + (int)An_sps->x.size()) < m_config.Threshold)
	{
		SampleSet_sps->x = Tp_sps->x;
		SampleSet_sps->x.insert(SampleSet_sps->x.end(), Tn_sps->x.begin(), Tn_sps->x.end());
		SampleSet_sps->x.insert(SampleSet_sps->x.end(), Unlabeled_sps->x.begin(), Unlabeled_sps->x.end());

		SampleSet_sps->yv = Tp_sps->yv;
		SampleSet_sps->yv.insert(SampleSet_sps->yv.end(), Tn_sps->yv.begin(), Tn_sps->yv.end());
		SampleSet_sps->yv.insert(SampleSet_sps->yv.end(), Unlabeled_sps->yv.begin(), Unlabeled_sps->yv.end());

		if (!m_config.quietMode && m_config.debugMode)
		{
			SampleSet_sps->images = Tp_sps->images;
			SampleSet_sps->images.insert(SampleSet_sps->images.end(), Tn_sps->images.begin(), Tn_sps->images.end());
			SampleSet_sps->images.insert(SampleSet_sps->images.end(), Unlabeled_sps->images.begin(), Unlabeled_sps->images.end());
		}

		flag = 0;
	}
	else
	{
		int repeat = (int)(n_Tposi / m_config.interval) + 1;
		SampleSet_sps->x = Ap_sps->x;
		SampleSet_sps->x.erase(SampleSet_sps->x.end() - repeat, SampleSet_sps->x.end());
		SampleSet_sps->x.insert(SampleSet_sps->x.end(), An_sps->x.begin(), An_sps->x.end() - repeat * 4);
		SampleSet_sps->x.insert(SampleSet_sps->x.end(), Tp_sps->x.begin(), Tp_sps->x.end());
		SampleSet_sps->x.insert(SampleSet_sps->x.end(), Tn_sps->x.begin(), Tn_sps->x.end());
		SampleSet_sps->x.insert(SampleSet_sps->x.end(), Unlabeled_sps->x.begin(), Unlabeled_sps->x.end());

		SampleSet_sps->yv = Ap_sps->yv;
		SampleSet_sps->yv.erase(SampleSet_sps->yv.end() - repeat, SampleSet_sps->yv.end());
		SampleSet_sps->yv.insert(SampleSet_sps->yv.end(), An_sps->yv.begin(), An_sps->yv.end() - repeat * 4);
		SampleSet_sps->yv.insert(SampleSet_sps->yv.end(), Tp_sps->yv.begin(), Tp_sps->yv.end());
		SampleSet_sps->yv.insert(SampleSet_sps->yv.end(), Tn_sps->yv.begin(), Tn_sps->yv.end());
		SampleSet_sps->yv.insert(SampleSet_sps->yv.end(), Unlabeled_sps->yv.begin(), Unlabeled_sps->yv.end());

		if (!m_config.quietMode && m_config.debugMode)
		{
			SampleSet_sps->images = Ap_sps->images;
			SampleSet_sps->images.erase(SampleSet_sps->images.end() - repeat, SampleSet_sps->images.end());
			SampleSet_sps->images.insert(SampleSet_sps->images.end(), An_sps->images.begin(), An_sps->images.end() - repeat * 4);
			SampleSet_sps->images.insert(SampleSet_sps->images.end(), Tp_sps->images.begin(), Tp_sps->images.end());
			SampleSet_sps->images.insert(SampleSet_sps->images.end(), Tn_sps->images.begin(), Tn_sps->images.end());
			SampleSet_sps->images.insert(SampleSet_sps->images.end(), Unlabeled_sps->images.begin(), Unlabeled_sps->images.end());
		}

		flag = 1;
	}
}

void GPR::costruKernel()
{
	int kersize((int)SampleSet_sps->x.size());
	if (m_kernel.BlockFlag())
	{
		m_K.block(0, 0, kersize, kersize) = ContextLogSimi0(SampleSet_sps->x);
	}
	else
	{
		for (int ind = 0; ind < kersize; ++ind)
		{
			for (int i = 0; i < ind; ++i)
			{
				m_K(i, ind) = m_kernel.Eval(SampleSet_sps->x[i], SampleSet_sps->x[ind]);
				m_K(ind, i) = m_K(i, ind);
			}
			m_K(ind, ind) = m_kernel.Eval(SampleSet_sps->x[ind]);
		}
	}
}

Eigen::MatrixXd GPR::ContextLogSimi0(const std::vector<Eigen::VectorXd> x)
{
	kNumBins = const_cast<Features&>(m_features).GetkNumBins();
	int n(x.size());
	int d((int)x[0].size());
	int numgrid((int)sqrt(d / kNumBins));
	MatrixXd Data = MatrixXd::Zero(kNumBins, n);
	MatrixXd Data2 = MatrixXd::Zero(n, n);
	MatrixXd Data_transpose = MatrixXd::Zero(n, kNumBins);
	MatrixXd Data2_transpose = MatrixXd::Zero(n, n);
	MatrixXd dist2 = MatrixXd::Zero(n, n);
	MatrixXd dist2ForSort = MatrixXd::Zero(n, n);
	double *pdist2;
	//Mat dist2cvtest = Mat(n, n, CV_64F);
	//CvMat pdist2cv;
	Mat sorted, indexkernel;
	VectorXd kNNdist2 = VectorXd::Zero(n);
	//VectorXd kNNdist2test = VectorXd::Zero(n);
	VectorXd localscale;
	MatrixXd localscale_transpose;
	MatrixXd LocalScale;
	MatrixXd Scontext = MatrixXd::Zero(n, n);
	double weight;
	int igrid, jgrid;
	float pha(4.9);
	double normalize(0.0);
	//struct timeb startTime1, endTime1, startTime2, endTime2;

#pragma omp parallel for private(pdist2,sorted,indexkernel,localscale,localscale_transpose,LocalScale,weight,igrid,jgrid) firstprivate(Data,Data_transpose,Data2,Data2_transpose,dist2,dist2ForSort,kNNdist2) reduction(+: normalize)
	for (int icell = 0; icell < numgrid*numgrid; ++icell)
	//for (int igrid = 0; igrid < numgrid; ++igrid)
	{
		jgrid = icell % numgrid;
		igrid = (int)((icell - jgrid)/numgrid);
		//for (int jgrid = 0; jgrid < numgrid; ++jgrid)
		//{
			
			for (int i = 0; i < n; ++i)
			{
				Data.col(i) = x[i].segment((igrid*numgrid + jgrid) * kNumBins, kNumBins);
#if VERBOSEMY3		
				cout << "Here is the matrix Data.col(i):" << endl << Data.col(i).transpose() << endl;
#endif
			}
#if VERBOSEMY3		
			cout << "Here is the matrix Data:" << endl << Data << endl;
#endif
			
			//VectorXd X2 = Data.cwiseProduct(Data).colwise().sum();
			for (int i = 0; i < n; ++i)
			{
				Data2.row(i) = Data.cwiseProduct(Data).colwise().sum();
			}
			Data2_transpose = Data2.transpose();
			Data_transpose = Data.transpose();
			dist2 = Data2 + Data2_transpose - 2 * Data_transpose*Data;
			//dist2 = Data2 + Data2.transpose() - 2 * Data.transpose()*Data;
#if VERBOSEMY3		
			cout << "Here is the matrix dist2:" << endl << dist2.block(200, 0, 188, 8) << endl;
#endif
			
			//for (int i = 0; i < n; ++i)
			//{
			//	for (int j = 0; j < n; ++j)
			//	{
					//float test = dist2(i, j);
			//		dist2cvtest.at<double>(i, j) = dist2(i, j);
			//		
			//	}
			//}
			dist2ForSort = dist2;
			pdist2 = (double *)(dist2ForSort.data());
			//Map<MatrixXd>(pdist2, dist2.rows(), dist2.cols()) = dist2;
			//cvInitMatHeader(&pdist2cv, n, n, CV_64FC1, pdist2);
			//Mat dist2cv(&pdist2cv, 1);
			//Mat dist2cv = Mat(n, n, CV_64F, pdist2);
			//sortIdx(dist2cv, indexkernel, CV_SORT_EVERY_COLUMN + CV_SORT_ASCENDING);
			//ftime(&startTime1);
			//cv::sort(dist2cv, sorted, CV_SORT_EVERY_COLUMN + CV_SORT_ASCENDING);
			//ftime(&endTime1);
			//ftime(&startTime1);
			for (int i = 0; i < n; ++i)
			{
				//kNNdist2(i) = dist2cvtest.at<float>(6, i);
				//kNNdist2(i) = CV_MAT_ELEM(pdist2cv, float, 6, i);
				//kNNdist2(i) = pdist2[i*n];
				//kNNdist2(i) = sorted.at<double>(6, i);
				kNNdist2(i) = GetMinKth(pdist2, i*n, (i+1)*n-1, 6); //GetMinKth will change the data that is pointed by pdist2, however this data will still be used in computing Scontext!!!!
				//cout << kNNdist2(i) << kNNdist2test(i) << endl;
			}
			//ftime(&endTime1);
#if VERBOSEMY4		
			cout << "Here is the matrix kNNdist2:" << endl << kNNdist2 << endl;
#endif
			//ftime(&startTime2);
			localscale = kNNdist2.cwiseSqrt();
#if VERBOSEMY3		
			cout << "Here is the matrix localscale:" << endl << localscale << endl;
#endif
			localscale_transpose = localscale.transpose();
			LocalScale = localscale*localscale_transpose;
			//LocalScale = localscale*localscale.transpose();
			weight = exp(-((igrid + 1.0 - numgrid / 2 - 0.5)*(igrid + 1.0 - numgrid / 2 - 0.5) + (jgrid + 1.0 - numgrid / 2 - 0.5)*(jgrid + 1.0 - numgrid / 2 - 0.5)) / (2 * pha * pha));
			Scontext += (-dist2).cwiseProduct(LocalScale.cwiseInverse()).array().exp().matrix()*weight;
			normalize += weight;
			/*Scontext += (-dist2).cwiseProduct(LocalScale.cwiseInverse()).array().exp().matrix();
			normalize += 1;*/
			//ftime(&endTime2);
			//cout << "ratio " << (float)((endTime1.time - startTime1.time) * 1000 + (endTime1.millitm - startTime1.millitm)) / ((endTime2.time - startTime2.time) * 1000 + (endTime2.millitm - startTime2.millitm)) << endl;
		//}
	}
	Scontext /= normalize;
	return Scontext;

}

/*Eigen::MatrixXd GPR::ContextLogSimi0(const std::vector<Eigen::VectorXd> x)
{
int n(x.size());
MatrixXd Data = MatrixXd::Zero(x[0].size(), n);
for (int i = 0; i < n; ++i)
{
Data.col(i) = x[i];
}
//VectorXd X2 = Data.cwiseProduct(Data).colwise().sum();
MatrixXd Data2 = MatrixXd::Zero(n, n);
for (int i = 0; i < n; ++i)
{
Data2.row(i) = Data.cwiseProduct(Data).colwise().sum();
}
MatrixXd dist2 = MatrixXd::Zero(n, n);
dist2 = Data2 + Data2.transpose() - 2 * Data.transpose()*Data;
Mat dist2cv = Mat(n, n, CV_32F);
for (int i = 0; i < n; ++i)
{
for (int j = 0; j < n; ++j)
{
dist2cv.at<float>(i, j) = dist2(i, j);
}
}
Mat sorted, indexkernel;

sortIdx(dist2cv, indexkernel, CV_SORT_EVERY_COLUMN + CV_SORT_ASCENDING);
cv::sort(dist2cv, sorted, CV_SORT_EVERY_COLUMN + CV_SORT_ASCENDING);

VectorXd kNNdist2 = VectorXd::Zero(n);
for (int i = 0; i < n; ++i)
{
kNNdist2(i) = sorted.at<float>(6, i);
}
VectorXd localscale(kNNdist2.cwiseSqrt());
MatrixXd LocalScale(localscale*localscale.transpose());
MatrixXd Scontext;
Scontext = (-dist2).cwiseProduct(LocalScale.cwiseInverse()).array().exp().matrix();
return Scontext;

}*/


void GPR::UpdateDebugImage()
{
	m_debugImage.setTo(0);

	int n = (int)Unlabeled_sps->x.size();

	if (n == 0) return;

	const int kCanvasSize = 600;
	int gridSize = (int)sqrtf((float)(n - 1)) + 1;
	int tileSize = (int)((float)kCanvasSize / gridSize);

	if (tileSize < 5)
	{
		cout << "too many support vectors to display" << endl;
		return;
	}

	Mat temp(tileSize, tileSize, CV_8UC1);
	int x = 0;
	int y = 0;
	int ind = 0;
	float vals[kMaxSVsGPR];
	memset(vals, 0, sizeof(float)*n);
	int drawOrder[kMaxSVsGPR];

	for (int i = 0; i < n; ++i)
	{

		drawOrder[ind] = i;
		vals[ind] = 1;
		++ind;

		Mat I = m_debugImage(cv::Rect(x, y, tileSize, tileSize));
		resize(Unlabeled_sps->images[Index.at<int>(i, 0)], temp, temp.size());
		cvtColor(temp, I, CV_GRAY2RGB);
		x += tileSize;
		if ((x + tileSize) > kCanvasSize)
		{
			y += tileSize;
			x = 0;
		}
	}

	n = 100;
	const int kKernelPixelSize = 2;
	int kernelSize = kKernelPixelSize*n;

	if (kernelSize < m_debugImage.cols && kernelSize < m_debugImage.rows)
	{
		Mat K = m_debugImage(cv::Rect(m_debugImage.cols - kernelSize, m_debugImage.rows - kernelSize, kernelSize, kernelSize));
		for (int i = 0; i < n; ++i)
		{
			for (int j = 0; j < n; ++j)
			{
				Mat Kij = K(cv::Rect(j*kKernelPixelSize, i*kKernelPixelSize, kKernelPixelSize, kKernelPixelSize));
				uchar v = (uchar)(255 * 0.5);
				Kij.setTo(Scalar(v, v, v));
			}
		}
	}
	else
	{
		kernelSize = 0;
	}


	Mat I = m_debugImage(cv::Rect(0, m_debugImage.rows - 200, m_debugImage.cols - kernelSize, 200));
	I.setTo(Scalar(255, 255, 255));
	IplImage II = I;
	setGraphColor(0);
	drawFloatGraph(vals, n, &II, 0.f, 0.f, I.cols, I.rows);
}


void GPR::UpdateDebugImage1()
{
	m_debugImage1.setTo(0);

	int n = (int)Unlabeled_sps->x.size();

	if (n == 0) return;

	const int kCanvasSize = 600;
	int gridSize = (int)sqrtf((float)(n - 1)) + 1;
	int tileSize = (int)((float)kCanvasSize / gridSize);

	if (tileSize < 5)
	{
		cout << "too many support vectors to display" << endl;
		return;
	}

	Mat temp(tileSize, tileSize, CV_8UC1);
	int x = 0;
	int y = 0;
	int ind = 0;
	float vals[kMaxSVsGPR];
	memset(vals, 0, sizeof(float)*n);
	int drawOrder[kMaxSVsGPR];

	for (int i = 0; i < n; ++i)
	{

		drawOrder[ind] = i;
		vals[ind] = 1;
		++ind;

		Mat I = m_debugImage1(cv::Rect(x, y, tileSize, tileSize));
		resize(Unlabeled_sps->images[Index1.at<int>(i, 0)], temp, temp.size());
		cvtColor(temp, I, CV_GRAY2RGB);
		x += tileSize;
		if ((x + tileSize) > kCanvasSize)
		{
			y += tileSize;
			x = 0;
		}
	}

	n = 100;
	const int kKernelPixelSize = 2;
	int kernelSize = kKernelPixelSize*n;

	if (kernelSize < m_debugImage1.cols && kernelSize < m_debugImage1.rows)
	{
		Mat K = m_debugImage1(cv::Rect(m_debugImage1.cols - kernelSize, m_debugImage1.rows - kernelSize, kernelSize, kernelSize));
		for (int i = 0; i < n; ++i)
		{
			for (int j = 0; j < n; ++j)
			{
				Mat Kij = K(cv::Rect(j*kKernelPixelSize, i*kKernelPixelSize, kKernelPixelSize, kKernelPixelSize));
				uchar v = (uchar)(255 * 0.5);
				Kij.setTo(Scalar(v, v, v));
			}
		}
	}
	else
	{
		kernelSize = 0;
	}


	Mat I = m_debugImage1(cv::Rect(0, m_debugImage1.rows - 200, m_debugImage1.cols - kernelSize, 200));
	I.setTo(Scalar(255, 255, 255));
	IplImage II = I;
	setGraphColor(0);
	drawFloatGraph(vals, n, &II, 0.f, 0.f, I.cols, I.rows);
}

void GPR::UpdateDebugImage2()
{
	m_debugImage2.setTo(0);

	int n = (int)SampleSet_sps->x.size();

	if (n == 0) return;

	const int kCanvasSize = 600;
	int gridSize = (int)sqrtf((float)(n - 1)) + 1;
	int tileSize = (int)((float)kCanvasSize / gridSize);

	if (tileSize < 5)
	{
		cout << "too many support vectors to display" << endl;
		return;
	}

	Mat temp(tileSize, tileSize, CV_8UC1);
	int x = 0;
	int y = 0;
	int ind = 0;
	float vals[kMaxSVsGPR];
	memset(vals, 0, sizeof(float)*n);
	int drawOrder[kMaxSVsGPR];

	for (int i = 0; i < n; ++i)
	{

		drawOrder[ind] = i;
		vals[ind] = 1;
		++ind;

		Mat I = m_debugImage2(cv::Rect(x, y, tileSize, tileSize));
		resize(SampleSet_sps->images[i], temp, temp.size());
		cvtColor(temp, I, CV_GRAY2RGB);
		x += tileSize;
		if ((x + tileSize) > kCanvasSize)
		{
			y += tileSize;
			x = 0;
		}
	}

	///n = 100;
	const int kKernelPixelSize = 2;
	int kernelSize = kKernelPixelSize*n;

	double kmin = m_K.minCoeff();
	double kmax = m_K.maxCoeff();

	if (kernelSize < m_debugImage2.cols && kernelSize < m_debugImage2.rows)
	{
		Mat K = m_debugImage2(cv::Rect(m_debugImage2.cols - kernelSize, m_debugImage2.rows - kernelSize, kernelSize, kernelSize));
		for (int i = 0; i < n; ++i)
		{
			for (int j = 0; j < n; ++j)
			{
				Mat Kij = K(cv::Rect(j*kKernelPixelSize, i*kKernelPixelSize, kKernelPixelSize, kKernelPixelSize));
				uchar v = (uchar)(255 * (m_K(drawOrder[i], drawOrder[j]) - kmin) / (kmax - kmin));
				Kij.setTo(Scalar(v, v, v));
			}
		}
	}
	else
	{
		kernelSize = 0;
	}


	Mat I = m_debugImage2(cv::Rect(0, m_debugImage2.rows - 200, m_debugImage2.cols - kernelSize, 200));
	I.setTo(Scalar(255, 255, 255));
	IplImage II = I;
	setGraphColor(0);
	drawFloatGraph(vals, n, &II, 0.f, 0.f, I.cols, I.rows);
}


double GPR::GetMinKth(double* ary, int lowIndex, int highIndex, int k)
{
	if (lowIndex == highIndex)
	{
		return ary[lowIndex];
	}

	if (lowIndex > highIndex)
	{
		if (highIndex < 0)
		{
			return ary[lowIndex];
		}
		return ary[highIndex];
	}
	int index = PartitionMin(ary, lowIndex, highIndex);
	int relativeIndex = index - lowIndex + 1;
	if (relativeIndex == k)
	{
		return ary[index];
	}
	if (relativeIndex > k)
	{
		return GetMinKth(ary, lowIndex, index - 1, k);
	}
	else
	{
		return GetMinKth(ary, index + 1, highIndex, k - relativeIndex);
	}

}

int GPR::PartitionMin(double* ary, int lowIndex, int highIndex)
{
	int small = lowIndex - 1;
	for (int index = lowIndex; index < highIndex; ++index)
	{
		if (ary[index] < ary[highIndex])
		{
			++small;
			if (small != index)
				Swap(ary[index], ary[small]);
		}
	}
	++small;
	Swap(ary[small], ary[highIndex]);
	return small;

	/*double middle = ary[lowIndex];
	int start = lowIndex;
	int end = highIndex;
	while (start < end)
	{
		while (start < end && ary[end] >= middle)
		{
			end--;
		}
		Swap(ary[start], ary[end]);

		while (start < end && ary[start] <= middle)
		{
			start++;
		}
		Swap(ary[start], ary[end]);
	}
	return start;*/
}

void GPR::Swap(double& x, double& y)
{
	double temp = x;
	x = y;
	y = temp;
}
