/*
* TGPR: Transfer Learning Based Visual Tracking with Gaussian Process Regression
*
* Code to accompany the paper:
*   Transfer Learning Based Visual Tracking with Gaussian Process Regression
*   Jin Gao, Haibin Ling, Weiming Hu, Junliang Xing
*   ECCV2014, Part III, LNC 8691, pp. 188-203, 2014
*
* Copyright (C) 2014 Jin Gao
*
* This file is part of TGPR.
*
*
*The whole framework of this code is based on the framework of the publicly available code:
*
* Struck: Structured Output Tracking with Kernels
*
* Struck is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* Struck is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with Struck.  If not, see <http://www.gnu.org/licenses/>.
*
*/

#include "HogFeatures.h"
#include "Config.h"
#include "Sample.h"
#include "Rect.h"
#define EIGEN_NO_DEBUG
#include <iostream>
#include <Eigen/Eigenvalues> 
#include <cmath>

using namespace Eigen;
using namespace cv;
using namespace std;

static const int numHogFeature = 9;
static const int Nummodes = 5;
static const int kNumBins = numHogFeature*Nummodes;
static const int kNumLevels = 1;

HogFeatures::HogFeatures(const Config& conf) :
kNumCellsX(conf.Nblock),
kNumCellsY(conf.Nblock),
hogtmp1(VectorXd::Zero(numHogFeature*kNumCellsX*kNumCellsY)),
hogtmp2(VectorXd::Zero(2*numHogFeature*kNumCellsX*kNumCellsY)),
hogtmp3(VectorXd::Zero(2*numHogFeature*kNumCellsX*kNumCellsY))
{
	int nc = 0;
	for (int i = 0; i < kNumLevels; ++i)
	{
		nc += (i + kNumCellsX)*(i + kNumCellsY);
	}
	SetCount(kNumBins*nc);
}

void HogFeatures::GetkNumBinsFeatures()
{
	kNumBinsFeatures = kNumBins;
}

void HogFeatures::UpdateFeatureVector(const Sample& s, Eigen::VectorXd& featVect)
{
	HOGDescriptor* desc;
	IntRect rect = s.GetROI(); // note this truncates to integers
	m_featVec.setZero();
	featVect = Eigen::VectorXd::Zero(m_featVec.size());
	//std::vector<float> hogtemp;
	int step = 0;
	for (int il = 0; il < kNumLevels; ++il)
	{
		int ncx = il + kNumCellsX;
		int ncy = il + kNumCellsY;
		int w = s.GetROI().Width() / ncx;
		int h = s.GetROI().Height() / ncy;
		cv::Rect roi1(rect.XMin(), rect.YMin(), rect.Width(), rect.Height());

		desc = new HOGDescriptor(cvSize(w, h), cvSize(w, h), cvSize(w, h), cvSize(w, h), numHogFeature);
		desc->compute(s.GetImage().GetImage(0)(roi1), hogdescript1, cvSize(w, h), cvSize(0, 0));
		for (int i = 0; i < hogdescript1.size(); ++i)
		{
			hogtmp1(i) = hogdescript1[i];
		}
		delete desc;

		if (w % 2 != 0)
		{
			desc = new HOGDescriptor(cvSize(w-1, h), cvSize((int)(w / 2), h), cvSize((int)(w / 2), h), cvSize((int)(w / 2), h), numHogFeature);
		}
		else
		{
			desc = new HOGDescriptor(cvSize(w, h), cvSize((int)(w / 2), h), cvSize((int)(w / 2), h), cvSize((int)(w / 2), h), numHogFeature);
		}	
		desc->compute(s.GetImage().GetImage(0)(roi1), hogdescript2, cvSize(w, h), cvSize(0, 0));
		for (int i = 0; i < hogdescript2.size(); ++i)
		{
			hogtmp2(i) = hogdescript2[i];
		}
		delete desc;

		if (h % 2 != 0)
		{
			desc = new HOGDescriptor(cvSize(w, h - 1), cvSize(w, (int)(h / 2)), cvSize(w, (int)(h / 2)), cvSize(w, (int)(h / 2)), numHogFeature);
		}
		else
		{
			desc = new HOGDescriptor(cvSize(w, h), cvSize(w, (int)(h / 2)), cvSize(w, (int)(h / 2)), cvSize(w, (int)(h / 2)), numHogFeature);
		}
		desc->compute(s.GetImage().GetImage(0)(roi1), hogdescript3, cvSize(w, h), cvSize(0, 0));
		for (int i = 0; i < hogdescript3.size(); ++i)
		{
			hogtmp3(i) = hogdescript3[i];
		}
		delete desc;

		for (int imode = 0; imode < Nummodes; ++imode)
		{
			for (int iy = 0; iy < ncy; ++iy)
			{
				for (int ix = 0; ix < ncx; ++ix)
				{
					++step;
					//hogtemp.assign(hogdescript.begin() + (iy*ncx + ix)*numHogFeature, hogdescript.begin() + step*numHogFeature);
					if (imode == 0)
					{
						featVect.segment((step - 1)*kNumBins + imode*numHogFeature, numHogFeature) = hogtmp1.segment((iy*ncx + ix)*numHogFeature, numHogFeature);
					}
					else if (imode == 1)
					{
						featVect.segment((step - 1)*kNumBins + imode*numHogFeature, numHogFeature) = hogtmp2.segment((iy*ncx + ix) * 2 * numHogFeature, numHogFeature);
					}
					else if (imode == 2)
					{
						featVect.segment((step - 1)*kNumBins + imode*numHogFeature, numHogFeature) = hogtmp2.segment((iy*ncx + ix) * 2 * numHogFeature + numHogFeature, numHogFeature);
					}
					else if (imode == 3)
					{
						featVect.segment((step - 1)*kNumBins + imode*numHogFeature, numHogFeature) = hogtmp3.segment((iy*ncx + ix + (iy % 2) * ncx) * numHogFeature, numHogFeature);
					}
					else if (imode == 4)
					{
						featVect.segment((step - 1)*kNumBins + imode*numHogFeature, numHogFeature) = hogtmp3.segment((iy*ncx + ix + ((iy % 2) + 1) * ncx) * numHogFeature, numHogFeature);
					}
				}
			}
			step = 0;
		}
	}
}