/*
* TGPR: Transfer Learning Based Visual Tracking with Gaussian Process Regression
*
* Code to accompany the paper:
*   Transfer Learning Based Visual Tracking with Gaussian Process Regression
*   Jin Gao, Haibin Ling, Weiming Hu, Junliang Xing
*   ECCV2014, Part III, LNC 8691, pp. 188-203, 2014
*
* Copyright (C) 2014 Jin Gao
*
* This file is part of TGPR.
*
*
*The whole framework of this code is based on the framework of the publicly available code:
*
* Struck: Structured Output Tracking with Kernels
*
* Struck is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* Struck is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with Struck.  If not, see <http://www.gnu.org/licenses/>.
*
*/

#include "ImageRep.h"

#include <cassert>

#include <opencv/highgui.h>
#include<cmath>

using namespace std;
using namespace cv;

static const int kNumBins = 16;
static const int numCovFeature = 6; //[x, y, I(x,y), Idx, Idy, sqrt(Idx^2+Idy^2)]

ImageRep::ImageRep(const Mat& image, bool computeIntegral, bool computeIntegralHist, bool computeIntegralCov, bool colour) :
m_channels(colour ? 3 : 1),
m_rect(0, 0, image.cols, image.rows)
{
	for (int i = 0; i < m_channels; ++i)
	{
		m_images.push_back(Mat(image.rows, image.cols, CV_8UC1));
		if (computeIntegral) m_integralImages.push_back(Mat(image.rows + 1, image.cols + 1, CV_32SC1));
		if (computeIntegralHist)
		{
			for (int j = 0; j < kNumBins; ++j)
			{
				m_integralHistImages.push_back(Mat(image.rows + 1, image.cols + 1, CV_32SC1));
			}
		}
		if (computeIntegralCov)
		{
			for (int j = 0; j < numCovFeature; ++j)
			{
				m_integralCovP.push_back(Mat(image.rows + 1, image.cols + 1, CV_64FC1));
				CovFeatureVector.push_back(Mat(image.rows, image.cols, CV_64FC1));
			}
			for (int j = 0; j < (numCovFeature*numCovFeature); ++j)
			{
				m_integralCovQ.push_back(Mat(image.rows + 1, image.cols + 1, CV_64FC1));
			}
		}
	}

	if (colour)
	{
		assert(image.channels() == 3);
		split(image, m_images);
	}
	else
	{
		assert(image.channels() == 1 || image.channels() == 3);
		if (image.channels() == 3)
		{
			cvtColor(image, m_images[0], CV_RGB2GRAY);
		}
		else if (image.channels() == 1)
		{
			image.copyTo(m_images[0]);
		}
	}

	if (computeIntegral)
	{
		for (int i = 0; i < m_channels; ++i)
		{
			//equalizeHist(m_images[i], m_images[i]);
			integral(m_images[i], m_integralImages[i]);
		}
	}

	if (computeIntegralHist)
	{
		Mat tmp(image.rows, image.cols, CV_8UC1);
		tmp.setTo(0);
		for (int j = 0; j < kNumBins; ++j)
		{
			for (int y = 0; y < image.rows; ++y)
			{
				const uchar* src = m_images[0].ptr(y);
				uchar* dst = tmp.ptr(y);
				for (int x = 0; x < image.cols; ++x)
				{
					int bin = (int)(((float)*src / 256)*kNumBins);
					*dst = (bin == j) ? 1 : 0;
					++src;
					++dst;
				}
			}

			integral(tmp, m_integralHistImages[j]);
		}
	}

	if (computeIntegralCov)
	{
		Mat tlt_1 = Mat(1, 3, CV_8SC1);
		Mat tlt_2 = Mat(3, 1, CV_8SC1);
		tlt_1.at<char>(0, 0) = -1;
		tlt_1.at<char>(0, 1) = 0;
		tlt_1.at<char>(0, 2) = 1;
		tlt_2.at<char>(0, 0) = -1;
		tlt_2.at<char>(1, 0) = 0;
		tlt_2.at<char>(2, 0) = 1;
		conv2(m_images[0], tlt_1, CONVOLUTION_SAME, CovFeatureVector[3]);
		conv2(m_images[0], tlt_2, CONVOLUTION_SAME, CovFeatureVector[4]);
		vector<double> testTemp;

		CovFeatureVector[3] = abs(CovFeatureVector[3]);
		CovFeatureVector[4] = abs(CovFeatureVector[4]);

		for (int i = 0; i < m_images[0].rows; ++i)
		{
			const uchar* src = m_images[0].ptr(i);
			for (int j = 0; j < m_images[0].cols; ++j)
			{
				CovFeatureVector[0].at<double>(i, j) = (double)i;
				CovFeatureVector[1].at<double>(i, j) = (double)j;
				CovFeatureVector[2].at<double>(i, j) = (double)*src;
				++src;
				CovFeatureVector[5].at<double>(i, j) = sqrt(CovFeatureVector[3].at<double>(i, j)*CovFeatureVector[3].at<double>(i, j) + CovFeatureVector[4].at<double>(i, j)*CovFeatureVector[4].at<double>(i, j));
			}
		}

		for (int i = 0; i < numCovFeature; ++i)
		{
			integral(CovFeatureVector[i], m_integralCovP[i]);
		}
		Mat dst;
		for (int i = 0; i < numCovFeature; ++i)
		{
			for (int ind = i; ind < numCovFeature; ++ind)
			{
				multiply(CovFeatureVector[i], CovFeatureVector[ind], dst);
				integral(dst, m_integralCovQ[i*numCovFeature + ind]);
				m_integralCovQ[ind*numCovFeature + i] = m_integralCovQ[i*numCovFeature + ind];
			}
		}

	}
}

int ImageRep::Sum(const IntRect& rRect, int channel) const
{
	assert(rRect.XMin() >= 0 && rRect.YMin() >= 0 && rRect.XMax() <= m_images[0].cols && rRect.YMax() <= m_images[0].rows);
	return m_integralImages[channel].at<int>(rRect.YMin(), rRect.XMin()) +
		m_integralImages[channel].at<int>(rRect.YMax(), rRect.XMax()) -
		m_integralImages[channel].at<int>(rRect.YMax(), rRect.XMin()) -
		m_integralImages[channel].at<int>(rRect.YMin(), rRect.XMax());
}

void ImageRep::Hist(const IntRect& rRect, Eigen::VectorXd& h) const
{
	assert(rRect.XMin() >= 0 && rRect.YMin() >= 0 && rRect.XMax() <= m_images[0].cols && rRect.YMax() <= m_images[0].rows);
	int norm = rRect.Area();
	for (int i = 0; i < kNumBins; ++i)
	{
		int sum = m_integralHistImages[i].at<int>(rRect.YMin(), rRect.XMin()) +
			m_integralHistImages[i].at<int>(rRect.YMax(), rRect.XMax()) -
			m_integralHistImages[i].at<int>(rRect.YMax(), rRect.XMin()) -
			m_integralHistImages[i].at<int>(rRect.YMin(), rRect.XMax());
		h[i] = (float)sum / norm;
	}
}

cv::Mat ImageRep::Cov(const IntRect& rRect) const
{
	//assert(rRect.XMin() >= 0 && rRect.YMin() >= 0 && rRect.XMax() <= m_images[0].cols && rRect.YMax() <= m_images[0].rows);
	Mat CovMatrix = Mat(numCovFeature, numCovFeature, CV_64FC1);
	int x1 = rRect.XMin();
	int y1 = rRect.YMin();
	int x2 = rRect.XMax();
	int y2 = rRect.YMax();
	if (x1 < 0)
	{
		x1 = 0;
	}
	if (x2 >= m_images[0].cols)
	{
		x2 = m_images[0].cols - 1;
	}
	if (y1 < 0)
	{
		y1 = 0;
	}
	if (y2 >= m_images[0].rows)
	{
		y2 = m_images[0].rows - 1;
	}
	int norm = (x2 - x1)*(y2 - y1);
	Mat dst;
	gemm(ToPVector(y2, x2) + ToPVector(y1, x1) - ToPVector(y1, x2) - ToPVector(y2, x1), ToPVector(y2, x2) + ToPVector(y1, x1) - ToPVector(y1, x2) - ToPVector(y2, x1), 1, NULL, 0, dst, CV_GEMM_B_T);
	CovMatrix = (1. / (norm - 1))*(ToQMatrix(y2, x2) + ToQMatrix(y1, x1) - ToQMatrix(y2, x1) - ToQMatrix(y1, x2) - (1. / norm)*dst);
	return CovMatrix;
}

cv::Mat ImageRep::ToPVector(const int x, const int y) const
{
	Mat PVector = Mat(numCovFeature, 1, CV_64FC1);
	for (int j = 0; j < numCovFeature; ++j)
	{
		PVector.at<double>(j, 0) = m_integralCovP[j].at<double>(x, y);
	}
	return PVector;
}

cv::Mat ImageRep::ToQMatrix(const int x, const int y) const
{
	Mat QMatrix = Mat(numCovFeature, numCovFeature, CV_64FC1);
	for (int i = 0; i < numCovFeature; ++i)
	{
		for (int ind = i; ind < numCovFeature; ++ind)
		{
			QMatrix.at<double>(i, ind) = m_integralCovQ[i*numCovFeature + ind].at<double>(x, y);
			QMatrix.at<double>(ind, i) = QMatrix.at<double>(i, ind);
		}
	}
	return QMatrix;
}

void ImageRep::conv2(const Mat &img, const Mat& kernel, ConvolutionType type, Mat& dest) const  //Convolution used for calculating the gradient
{
	Mat source = img;
	Mat kernelT;
	if (CONVOLUTION_FULL == type)
	{
		source = Mat();
		const int additionalRows = kernel.rows - 1, additionalCols = kernel.cols - 1;
		copyMakeBorder(img, source, (additionalRows + 1) / 2, additionalRows / 2, (additionalCols + 1) / 2, additionalCols / 2, BORDER_CONSTANT, Scalar(0));
	}
	Point anchor(kernel.cols - kernel.cols / 2 - 1, kernel.rows - kernel.rows / 2 - 1);
	int borderMode = BORDER_CONSTANT;
	flip(kernel, kernelT, -1);
	filter2D(source, dest, CV_64FC1, kernelT, anchor, 0, borderMode);
	if (CONVOLUTION_VALID == type)
	{
		dest = dest.colRange((kernel.cols - 1) / 2, dest.cols - kernel.cols / 2).rowRange((kernel.rows - 1) / 2, dest.rows - kernel.rows / 2);
	}
}
