/*
* TGPR: Transfer Learning Based Visual Tracking with Gaussian Process Regression
*
* Code to accompany the paper:
*   Transfer Learning Based Visual Tracking with Gaussian Process Regression
*   Jin Gao, Haibin Ling, Weiming Hu, Junliang Xing
*   ECCV2014, Part III, LNC 8691, pp. 188-203, 2014
*
* Copyright (C) 2014 Jin Gao
*
* This file is part of TGPR.
*
*
*The whole framework of this code is based on the framework of the publicly available code:
*
* Struck: Structured Output Tracking with Kernels
*
* Struck is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* Struck is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with Struck.  If not, see <http://www.gnu.org/licenses/>.
*
*/

#include "Sampler.h"
#include "Config.h"
#define EIGEN_NO_DEBUG
#include <Eigen/Core>

#include <opencv/cv.h>
#include <opencv/highgui.h>

#define _USE_MATH_DEFINES
#include <cmath>
#include <math.h>

using namespace cv;
using namespace std;
using namespace Eigen;

/*vector<FloatRect> Sampler::OffsetSamples2(FloatRect centre, int grid)
{
vector<FloatRect> samples;

FloatRect s(centre);
float rstep = centre.Width() / (2.5*(float)grid);
float cstep = centre.Height() / (2.5*(float)grid);
samples.push_back(s);

int r1 = (1.3 * grid) * (1.3 * grid);
int r2 = (2 * grid - 1) * (2 * grid - 1);
//int r2 = (1.8 * grid) * (1.8 * grid);
for (int iy = -(2 * grid - 1); iy <= (2 * grid - 1); ++iy)
{
for (int ix = -(2 * grid - 1); ix <= (2 * grid - 1); ++ix)
{
if (ix*ix + iy*iy < r1) continue;
if (ix*ix + iy*iy > r2) continue;
//if (iy == 0 && ix == 0) continue; // already put this one at the start

float x = centre.XMin() + ix*rstep;
float y = centre.YMin() + iy*cstep;
if (ix % 2 != 0 || iy % 2 != 0) continue;
s.SetXMin(x);
s.SetYMin(y);
samples.push_back(s);
}
}

return samples;
}*/

vector<FloatRect> Sampler::OffsetSamples2(FloatRect centre, int grid)
{
	vector<FloatRect> samples, temp;
	int NegNum = 160;
	int Num;

	FloatRect s(centre);
	float rstep = centre.Width() / (2.5*(float)grid);
	float cstep = centre.Height() / (2.5*(float)grid);
	samples.push_back(s);

	int r1 = (1.0 * grid) * (1.0 * grid);
	int r2 = (2.5 * grid - 1) * (2.5 * grid - 1);
	//int r2 = (1.8 * grid) * (1.8 * grid);
	for (int iy = -(2.5 * grid - 1); iy <= (2.5 * grid - 1); ++iy)
	{
		for (int ix = -(2.5 * grid - 1); ix <= (2.5 * grid - 1); ++ix)
		{
			if (ix*ix + iy*iy < r1) continue;
			if (ix*ix + iy*iy > r2) continue;

			float x = centre.XMin() + ix*rstep;
			float y = centre.YMin() + iy*cstep;

			s.SetXMin(x);
			s.SetYMin(y);
			temp.push_back(s);
		}
	}
	int n = (int)temp.size();
	int *Index = new int[n];
	randperm(n, Index);
	if (n < NegNum)
	{
		Num = n;
	}
	else
	{
		Num = NegNum;
	}
	for (int j = 0; j < Num; j++)
	{
		samples.push_back(temp[Index[j]]);
	}
	delete[] Index;

	return samples;
}

vector<FloatRect> Sampler::OffsetSamples(FloatRect centre, float grid)
{
	vector<FloatRect> samples;

	FloatRect s(centre);
	float rstep = centre.Width() / grid;
	float cstep = centre.Height() / grid;
	samples.push_back(s);

	int r2 = 1 * 1;
	for (int iy = -1; iy <= 1; ++iy)
	{
		for (int ix = -1; ix <= 1; ++ix)
		{
			if (ix*ix + iy*iy > r2) continue;
			if (iy == 0 && ix == 0) continue; // already put this one at the start

			float x = centre.XMin() + ix*rstep;
			float y = centre.YMin() + iy*cstep;

			s.SetXMin(x);
			s.SetYMin(y);
			samples.push_back(s);
		}
	}

	return samples;
}

vector<FloatRect> Sampler::UnlabeledSamples(FloatRect centre, float widthinit, float heightinit, int Nparticle, float affisigx_r, float affisigy_r, float affisigscale_r, bool flagcase)
{
	int Nparticle_c(Nparticle);
	if (flagcase)
	{
		Nparticle_c = (int)(1.0*Nparticle_c);
	}
	vector<FloatRect> samples;
	Mat randgaussian1(1, Nparticle_c, CV_32F);
	randn(randgaussian1, Scalar::all(0.0), Scalar::all(1.0));

	Mat randgaussian2(1, Nparticle_c, CV_32F);
	randn(randgaussian2, Scalar::all(0.0), Scalar::all(1.0));

	Mat randgaussian3(1, Nparticle_c, CV_32F);
	randn(randgaussian3, Scalar::all(0.2), Scalar::all(1.0));

	Mat randgaussian4(1, Nparticle_c, CV_32F);
	randn(randgaussian4, Scalar::all(0.0), Scalar::all(1.0));

	FloatRect s(centre);
	float affisigx = centre.Width() / 4;
	if ((affisigx - affisigx_r) > 1e-10)
	{
		affisigx = affisigx_r;
	}
	else if ((affisigx - 8.0) < 1e-10)
	{
		affisigx = 8.0;
	}
	else
	{

	}
	float affisigy = centre.Height() / 4;
	if ((affisigy - affisigy_r) > 1e-10)
	{
		affisigy = affisigy_r;
	}
	else if ((affisigy - 8.0) < 1e-10)
	{
		affisigy = 8.0;
	}
	else
	{

	}
	if (flagcase)
	{
		affisigx *= 1.0;
		affisigy *= 1.0;
	}
	float affisigw = affisigscale_r;
	//float affisigw = 0.0f;
	float affisighw = 0.0f;
	samples.push_back(s);

	for (int in = 0; in < (Nparticle_c - 1); ++in)
	{
		float x = centre.XMin() + randgaussian1.at<float>(0, in)*affisigx;
		float y = centre.YMin() + randgaussian2.at<float>(0, in)*affisigy;
		float w = (centre.Width() / 64 + randgaussian3.at<float>(0, in)*affisigw) * 64;
		float h = (heightinit / widthinit + randgaussian4.at<float>(0, in)*affisighw) * w;

		s.SetXMin(x);
		s.SetYMin(y);
		s.SetWidth(w);
		s.SetHeight(h);
		samples.push_back(s);
	}

	return samples;
}