/*
* TGPR: Transfer Learning Based Visual Tracking with Gaussian Process Regression
*
* Code to accompany the paper:
*   Transfer Learning Based Visual Tracking with Gaussian Process Regression
*   Jin Gao, Haibin Ling, Weiming Hu, Junliang Xing
*   ECCV2014, Part III, LNC 8691, pp. 188-203, 2014
*
* Copyright (C) 2014 Jin Gao
*
* This file is part of TGPR.
*
*
*The whole framework of this code is based on the framework of the publicly available code:
*
* Struck: Structured Output Tracking with Kernels
*
* Struck is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* Struck is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with Struck.  If not, see <http://www.gnu.org/licenses/>.
*
*/

#include "Tracker.h"
#include "Config.h"
#include "ImageRep.h"
#include "Sampler.h"
#include "Sample.h"
#include "GraphUtils/GraphUtils.h"
#include "Timing.h"

#include "CovFeatures.h"
#include "HogFeatures.h"
#include "HaarFeatures.h"
#include "RawFeatures.h"
#include "HistogramFeatures.h"
#include "MultiFeatures.h"

#include "Kernels.h"

#include "GPR.h"

#include <opencv/cv.h>
#include <opencv/highgui.h>
#define EIGEN_NO_DEBUG
#include <Eigen/Core>

#include <vector>
#include <algorithm>
#include <sys/timeb.h>
#include <ctime>
#include <climits>

using namespace cv;
using namespace std;
using namespace Eigen;

Tracker::Tracker(const Config& conf, Timing& Timer) :
m_config(conf),
m_pTiming(&Timer),
m_initialised(false),
m_pclassifier(0),
m_needsIntegralImage(false),
m_needsIntegralCov(false),
flagcase(false)
{
	Reset();
}

Tracker::~Tracker()
{
	delete m_pclassifier;
	for (int i = 0; i < (int)m_features.size(); ++i)
	{
		delete m_features[i];
		delete m_kernels[i];
	}
}

void Tracker::Reset()
{
	m_initialised = false;
	if (m_pclassifier) delete m_pclassifier;
	for (int i = 0; i < (int)m_features.size(); ++i)
	{
		delete m_features[i];
		delete m_kernels[i];
	}
	m_features.clear();
	m_kernels.clear();

	m_needsIntegralImage = false;
	m_needsIntegralHist = false;
	m_needsIntegralCov = false;
	flagcase = false;

	int numFeatures = m_config.features.size();
	vector<int> featureCounts;
	for (int i = 0; i < numFeatures; ++i)
	{
		switch (m_config.features[i].feature)
		{
		case Config::kFeatureTypeCovariance:
			m_features.push_back(new CovFeatures(m_config));
			m_needsIntegralCov = true;
			break;
		case Config::kFeatureTypeHog:
			m_features.push_back(new HogFeatures(m_config));
			break;
		case Config::kFeatureTypeHaar:
			m_features.push_back(new HaarFeatures(m_config));
			m_needsIntegralImage = true;
			break;
		case Config::kFeatureTypeRaw:
			m_features.push_back(new RawFeatures(m_config));
			break;
		case Config::kFeatureTypeHistogram:
			m_features.push_back(new HistogramFeatures(m_config));
			m_needsIntegralHist = true;
			break;
		}
		featureCounts.push_back(m_features.back()->GetCount());

		switch (m_config.features[i].kernel)
		{
		case Config::kKernelTypeBlocksimilarity:
			m_kernels.push_back(new BlocksimilarityKernel());
			break;
		case Config::kKernelTypeLinear:
			m_kernels.push_back(new LinearKernel());
			break;
		case Config::kKernelTypeGaussian:
			m_kernels.push_back(new GaussianKernel(m_config.features[i].params[0]));
			break;
		case Config::kKernelTypeIntersection:
			m_kernels.push_back(new IntersectionKernel());
			break;
		case Config::kKernelTypeChi2:
			m_kernels.push_back(new Chi2Kernel());
			break;
		}
	}

	if (numFeatures > 1)
	{
		MultiFeatures* f = new MultiFeatures(m_features);
		m_features.push_back(f);

		MultiKernel* k = new MultiKernel(m_kernels, featureCounts);
		m_kernels.push_back(k);
	}

	m_pclassifier = new GPR(m_config, *m_features.back(), *m_kernels.back(), *m_pTiming);
}

void Tracker::InitialiseGPR(const cv::Mat& frame, FloatRect bb, const int& frameInd)
{
	m_bb = IntRect(bb); //Bbox of initial frame: upleft x, upleft y, width, height
	width_initial = m_bb.Width(); //storing width of Bbox in the initial frame
	height_initial = m_bb.Height(); //storing height of Bbox in the initial frame; These two values are used in Sampler::UnlabeledSamples().
	ImageRep image(frame, m_needsIntegralImage, m_needsIntegralHist, m_needsIntegralCov); //construct the integral image using different features
	for (int i = 0; i < 1; ++i)
	{
		UpdateSampleSet(image, frameInd); //Collection of AuxiliaryTarget samples
	}
	m_initialised = true;
}

FloatRect Tracker::meanRect(const std::vector<FloatRect>& keptRects, const std::vector<int>& FIVE)  //the average over locations of the best selected samples
{
	int top = FIVE.size();
	FloatRect meanRec;
	float xmin = 0.f;
	float ymin = 0.f;
	float width = 0.f;
	float height = 0.f;
	for (int i = 0; i < top; ++i)
	{
		xmin += keptRects[FIVE[i]].XMin();
		ymin += keptRects[FIVE[i]].YMin();
		width += keptRects[FIVE[i]].Width();
		height += keptRects[FIVE[i]].Height();
	}
	xmin = xmin / top;
	ymin = ymin / top;
	width = width / top;
	height = height / top;
	//width = keptRects[FIVE[0]].Width();
	//height = keptRects[FIVE[0]].Height();
	meanRec = FloatRect(xmin, ymin, width, height);
	//meanRec = FloatRect(keptRects[FIVE[0]].XMin(), keptRects[FIVE[0]].YMin(), keptRects[FIVE[0]].Width(), keptRects[FIVE[0]].Height());
	return meanRec;
}

void Tracker::Track(const cv::Mat& frame, const int& frameInd)
{
	assert(m_initialised);
	ftime(&(m_pTiming->startTimer1));
	ImageRep image(frame, m_needsIntegralImage, m_needsIntegralHist, m_needsIntegralCov); //construct the integral image using different features
	ftime(&(m_pTiming->endTimer1));
	m_pTiming->storetime(m_pTiming->TimeStore1, m_pTiming->startTimer1, m_pTiming->endTimer1);

	vector<FloatRect> rects = Sampler::UnlabeledSamples(m_bb, width_initial, height_initial, m_config.Nparticle, m_config.affisigx, m_config.affisigy, m_config.affisigscale, flagcase); //sample unlabeled samples
	vector<FloatRect> keptRects;
	keptRects.reserve(rects.size());
	for (int i = 0; i < (int)rects.size(); ++i)
	{
		if (!rects[i].IsInside(image.GetRect())) continue; //Excluding the samples not in the view
		keptRects.push_back(rects[i]);
	}

	MultiSample sample(image, keptRects);
	vector<int> FIVE;
	FIVE.erase(FIVE.begin(), FIVE.end());
	ftime(&(m_pTiming->startTimer2));
	m_pclassifier->Eval(sample, FIVE, flagcase);   //Tracking Algorithm
	ftime(&(m_pTiming->endTimer2));
	m_pTiming->storetime(m_pTiming->TimeStore2, m_pTiming->startTimer2, m_pTiming->endTimer2);
	m_bb = meanRect(keptRects, FIVE); //the average location
	m_bb = IntRect(m_bb);

	ftime(&(m_pTiming->startTimer7));
	UpdateSampleSet(image, frameInd); ////updating of AuxiliaryTarget samples
	ftime(&(m_pTiming->endTimer7));
	m_pTiming->storetime(m_pTiming->TimeStore7, m_pTiming->startTimer7, m_pTiming->endTimer7);
#if VERBOSE		
	cout << "track score: " << bestScore << endl;
#endif
}

void Tracker::Debug()
{
	m_pclassifier->Debug();
}

void Tracker::UpdateSampleSet(const ImageRep& image, const int& frameInd)
{
	if (frameInd == 1 && !m_initialised)
	{
		// note these return the centre sample at index 0
		vector<FloatRect> rects = Sampler::OffsetSamples2(m_bb, 8); //collection of Target samples

		vector<FloatRect> keptRectsT;
		keptRectsT.push_back(rects[0]); // the true sample
		for (int i = 1; i < (int)rects.size(); ++i)
		{
			if (!rects[i].IsInside(image.GetRect())) continue;
			keptRectsT.push_back(rects[i]);
		}

#if VERBOSE		
		cout << keptRectsT.size() << " samples" << endl;
#endif

		MultiSample sampleT(image, keptRectsT);
		m_pclassifier->UpdateT(sampleT, 0); //feature extraction

		// note these return the centre sample at index 0
		rects = Sampler::OffsetSamples(m_bb, 2.5);  //collection of Auxiliary samples

		vector<FloatRect> keptRectsA;
		keptRectsA.push_back(rects[0]); // the true sample
		for (int i = 1; i < (int)rects.size(); ++i)
		{
			if (!rects[i].IsInside(image.GetRect())) continue;
			keptRectsA.push_back(rects[i]);
		}

#if VERBOSE		
		cout << keptRectsA.size() << " samples" << endl;
#endif

		MultiSample sampleA(image, keptRectsA);
		m_pclassifier->UpdateA(sampleA, 0); //feature extraction
	}
	else if ((frameInd % (m_config.interval)) == 0)  // Auxiliary Samples collected at intervals
	{
		// note these return the centre sample at index 0
		vector<FloatRect> rects = Sampler::OffsetSamples2(m_bb, 8);

		vector<FloatRect> keptRectsT;
		keptRectsT.push_back(rects[0]); // the true sample
		for (int i = 1; i < (int)rects.size(); ++i)
		{
			if (!rects[i].IsInside(image.GetRect())) continue;
			keptRectsT.push_back(rects[i]);
		}

#if VERBOSE		
		cout << keptRectsT.size() << " samples" << endl;
#endif

		MultiSample sampleT(image, keptRectsT);
		m_pclassifier->UpdateT(sampleT, 0);

		// note these return the centre sample at index 0
		rects = Sampler::OffsetSamples(m_bb, 2.5);

		vector<FloatRect> keptRectsA;
		keptRectsA.push_back(rects[0]); // the true sample
		for (int i = 1; i < (int)rects.size(); ++i)
		{
			if (!rects[i].IsInside(image.GetRect())) continue;
			keptRectsA.push_back(rects[i]);
		}

#if VERBOSE		
		cout << keptRectsA.size() << " samples" << endl;
#endif

		MultiSample sampleA(image, keptRectsA);
		m_pclassifier->UpdateA(sampleA, 0);
	}
	else if (frameInd > 1)  //Target Samples collected at every frame
	{
		// note these return the centre sample at index 0
		vector<FloatRect> rects = Sampler::OffsetSamples2(m_bb, 8);

		vector<FloatRect> keptRectsT;
		keptRectsT.push_back(rects[0]); // the true sample
		for (int i = 1; i < (int)rects.size(); ++i)
		{
			if (!rects[i].IsInside(image.GetRect())) continue;
			keptRectsT.push_back(rects[i]);
		}

#if VERBOSE		
		cout << keptRectsT.size() << " samples" << endl;
#endif

		MultiSample sampleT(image, keptRectsT);
		m_pclassifier->UpdateT(sampleT, 0);
	}
}
