/*
* TGPR: Transfer Learning Based Visual Tracking with Gaussian Process Regression
*
* Code to accompany the paper:
*   Transfer Learning Based Visual Tracking with Gaussian Process Regression
*   Jin Gao, Haibin Ling, Weiming Hu, Junliang Xing
*   ECCV2014, Part III, LNC 8691, pp. 188-203, 2014
*
* Copyright (C) 2014 Jin Gao
*
* This file is part of TGPR.
*
*
*The whole framework of this code is based on the framework of the publicly available code:
*
* Struck: Structured Output Tracking with Kernels
*
* Struck is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* Struck is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with Struck.  If not, see <http://www.gnu.org/licenses/>.
*
*/

#include "Tracker.h"
#include "Config.h"
#include "Timing.h"

#include <iostream>
#include <fstream>

#include <opencv/cv.h>
#include <opencv/highgui.h>
#include <sys/timeb.h>
#include <ctime>
#include <climits>


#pragma comment(lib,"IlmImf.lib")
#pragma comment(lib,"libjasper.lib")
#pragma comment(lib,"libjpeg.lib")
#pragma comment(lib,"libpng.lib")
#pragma comment(lib,"libtiff.lib")
#pragma comment(lib,"zlib.lib")

#pragma comment(lib,"opencv_calib3d248.lib")
#pragma comment(lib,"opencv_contrib248.lib")
#pragma comment(lib,"opencv_core248.lib")
#pragma comment(lib,"opencv_features2d248.lib")
#pragma comment(lib,"opencv_flann248.lib")
#pragma comment(lib,"opencv_gpu248.lib")
//#pragma comment(lib,"opencv_haartraining_engine.lib")
#pragma comment(lib,"opencv_highgui248.lib")
#pragma comment(lib,"opencv_imgproc248.lib")
#pragma comment(lib,"opencv_legacy248.lib")
#pragma comment(lib,"opencv_ml248.lib")
#pragma comment(lib,"opencv_nonfree248.lib")
#pragma comment(lib,"opencv_objdetect248.lib")
#pragma comment(lib,"opencv_ocl248.lib")
#pragma comment(lib,"opencv_photo248.lib")
#pragma comment(lib,"opencv_stitching248.lib")
#pragma comment(lib,"opencv_superres248.lib")
#pragma comment(lib,"opencv_ts248.lib")
#pragma comment(lib,"opencv_video248.lib")
#pragma comment(lib,"opencv_videostab248.lib")

#pragma comment(lib,"vfw32.lib")
#pragma comment(lib,"comctl32.lib")


using namespace std;
using namespace cv;

static const int kLiveBoxWidth = 80;
static const int kLiveBoxHeight = 80;

void rectangle(Mat& rMat, const FloatRect& rRect, const Scalar& rColour)
{
	IntRect r(rRect);
	rectangle(rMat, Point(r.XMin(), r.YMin()), Point(r.XMax(), r.YMax()), rColour);
}

int main(int argc, char* argv[])
{
	//Eigen::initParallel();
	Eigen::setNbThreads(1);
	// read config file
	string configPath = "config.txt";
	//	if (argc > 1)
	//	{
	//		configPath = argv[1];
	//	}
	Config conf(configPath); //read the parameter configuration 

	conf.Nblock = (argc > 12 ? atoi(argv[1]) : 4);
	conf.Nparticle = (argc > 13 ? atoi(argv[2]) : 200);
	conf.Threshold = (argc > 14 ? atoi(argv[3]) : 30);
	conf.interval = (argc > 15 ? atoi(argv[4]) : 6);
	conf.affisigx = (argc > 16 ? atof(argv[5]) : 14.0);
	conf.affisigy = (argc > 17 ? atof(argv[6]) : 14.0);
	conf.affisigscale = (argc > 18 ? atof(argv[7]) : 0.05);

	string FeatureName = (argc > 19 ? argv[8] : "covariance");
	if (FeatureName == "covariance") conf.features[0].feature = conf.kFeatureTypeCovariance;
	else if (FeatureName == "hog") conf.features[0].feature = conf.kFeatureTypeHog;
	else if (FeatureName == "raw") conf.features[0].feature = conf.kFeatureTypeRaw;
	else if (FeatureName == "haar") conf.features[0].feature = conf.kFeatureTypeHaar;
	else if (FeatureName == "histogram") conf.features[0].feature = conf.kFeatureTypeHistogram;
	else
	{
		cout << "error: unrecognised feature: " << FeatureName << endl;
	}

	string KernelName = (argc > 20 ? argv[9] : "blocksimilarity");
	if (KernelName == "blocksimilarity") conf.features[0].kernel = conf.kKernelTypeBlocksimilarity;
	else if (KernelName == "linear") conf.features[0].kernel = conf.kKernelTypeLinear;
	else if (KernelName == "intersection") conf.features[0].kernel = conf.kKernelTypeIntersection;
	else if (KernelName == "chi2") conf.features[0].kernel = conf.kKernelTypeChi2;
	else if (KernelName == "gaussian")
	{
		conf.features[0].params[0] = 0.2;
		conf.features[0].kernel = conf.kKernelTypeGaussian;
	}
	else
	{
		cout << "error: unrecognised kernel: " << KernelName << endl;
	}

	conf.debugMode = (bool)(argc > 21 ? atoi(argv[10]) : 0);
	conf.quietMode = (bool)(argc > 22 ? atoi(argv[11]) : 0);
	conf.sequenceName = argv[12];
	conf.resultsPath = argv[13];
	conf.sequenceBasePath = argv[14];
	string searchString("\\");
	string replaceString("/");
	string::size_type pos = 0;

	while ((pos = conf.sequenceBasePath.find(searchString, pos)) != string::npos) {
		conf.sequenceBasePath.replace(pos, searchString.size(), replaceString);
		pos++;
	}
	string::size_type pos1 = 0;

	while ((pos1 = conf.resultsPath.find(searchString, pos1)) != string::npos) {
		conf.resultsPath.replace(pos1, searchString.size(), replaceString);
		pos1++;
	}

	Timing Timer;
	//cout << conf << endl;

	if (conf.features.size() == 0)
	{
		cout << "error: no features specified in config" << endl;
		return EXIT_FAILURE;
	}

	ofstream outFile;
	if (conf.resultsPath != "")
	{
		string logFormat = "TGPRresults_txt/" + conf.sequenceName + "_TGPR.txt";
		outFile.open(logFormat.c_str(), ios::out);
		if (!outFile)
		{
			cout << "error: could not open results file: " << conf.resultsPath << endl;
			return EXIT_FAILURE;
		}
	}

	// if no sequence specified then use the camera
	bool useCamera = (conf.sequenceName == "");

	VideoCapture cap;

	int startFrame = -1;
	int endFrame = -1;
	FloatRect initBB;
	string imgFormat;
	float scaleW = 1.f;
	float scaleH = 1.f;
	struct timeb startTime, endTime;

	if (useCamera)
	{
		if (!cap.open(0))
		{
			cout << "error: could not start camera capture" << endl;
			return EXIT_FAILURE;
		}
		startFrame = 0;
		endFrame = INT_MAX;
		Mat tmp;
		cap >> tmp;
		scaleW = (float)conf.frameWidth / tmp.cols;
		scaleH = (float)conf.frameHeight / tmp.rows;

		initBB = IntRect(conf.frameWidth / 2 - kLiveBoxWidth / 2, conf.frameHeight / 2 - kLiveBoxHeight / 2, kLiveBoxWidth, kLiveBoxHeight);
		cout << "press 'i' to initialise tracker" << endl;
	}
	else
	{
		// parse frames file
		//string framesFilePath = conf.sequenceBasePath + "/" + conf.sequenceName + "/" + conf.sequenceName + "_frames.txt";
		//ifstream framesFile(framesFilePath.c_str(), ios::in);
		//if (!framesFile)
		//{
		//	cout << "error: could not open sequence frames file: " << framesFilePath << endl;
		//	return EXIT_FAILURE;
		//}
		//string framesLine;
		//getline(framesFile, framesLine);
		//sscanf(framesLine.c_str(), "%d,%d", &startFrame, &endFrame);
		//if (framesFile.fail() || startFrame == -1 || endFrame == -1)
		//{
		//	cout << "error: could not parse sequence frames file" << endl;
		//	return EXIT_FAILURE;
		//}
		startFrame = atoi(argv[15]);
		endFrame = atoi(argv[16]);

		imgFormat = conf.sequenceBasePath + "%0" + argv[17] + "d." + argv[18];

		// read first frame to get size
		char imgPath[256];
		sprintf(imgPath, imgFormat.c_str(), startFrame);
		//cout << imgPath << endl;
		Mat tmp = cv::imread(imgPath, 0); //0 indicates gray chanal
		/*if (tmp.rows < 540)
		{
		conf.frameHeight = 540;
		scaleH = (float)conf.frameHeight / tmp.rows;
		scaleW = scaleH;
		conf.frameWidth = scaleW*tmp.cols;
		}
		else
		{
		conf.frameHeight = tmp.rows;
		conf.frameWidth = tmp.cols;
		}
		cout << conf << endl;*/
		//imshow("resulttmp", tmp);
		//scaleW = 1.f;
		//scaleH = 1.f;

		// read init box from ground truth file
		//string gtFilePath = conf.sequenceBasePath + "/" + conf.sequenceName + "/" + conf.sequenceName + "_gt.txt";
		//ifstream gtFile(gtFilePath.c_str(), ios::in);
		//if (!gtFile)
		//{
		//	cout << "error: could not open sequence gt file: " << gtFilePath << endl;
		//	return EXIT_FAILURE;
		//}
		//string gtLine;
		//getline(gtFile, gtLine);
		float xmin = atof(argv[19]);
		float ymin = atof(argv[20]);
		float width = atof(argv[21]);
		float height = atof(argv[22]);

		if (width < 28 || height < 28)
		{
			scaleH = 1.5;
			scaleW = scaleH;
			conf.frameHeight = scaleH*tmp.rows;
			conf.frameWidth = scaleW*tmp.cols;
		}
		else
		{
			/*scaleH = 1.5;
			scaleW = scaleH;
			conf.frameHeight = scaleH*tmp.rows;
			conf.frameWidth = scaleW*tmp.cols;*/
			conf.frameHeight = tmp.rows;
			conf.frameWidth = tmp.cols;
		}
		cout << conf << endl;

		//sscanf(gtLine.c_str(), "%f,%f,%f,%f", &xmin, &ymin, &width, &height);
		//if (gtFile.fail() || xmin < 0.f || ymin < 0.f || width < 0.f || height < 0.f)
		if (xmin < 0.f || ymin < 0.f || width < 0.f || height < 0.f)
		{
			cout << "error: could not parse sequence gt file" << endl;
			return EXIT_FAILURE;
		}
		initBB = FloatRect(xmin*scaleW, ymin*scaleH, width*scaleW, height*scaleH);
	}



	Tracker tracker(conf, Timer);
	if (!conf.quietMode)
	{
		namedWindow("result");
	}
	///////////////////////////////////////////////////////
	//SaveResults_2
	Mat result(conf.frameHeight, conf.frameWidth, CV_8UC3);
	///////////////////////////////////////////////////////
	bool paused = false;
	bool doInitialise = false;

	///////////////////////////////////////////////////////
	//SaveResults_1
	//std::vector<cv::Mat> Temp;
	///////////////////////////////////////////////////////

	///////////////////////////////////////////////////////
	//SaveResults_2
	std::vector<FloatRect> Temp;
	///////////////////////////////////////////////////////

	ftime(&startTime);
	for (int frameInd = startFrame; frameInd <= endFrame; ++frameInd)
	{
		///////////////////////////////////////////////////////
		//SaveResults_1
		//Mat result(conf.frameHeight, conf.frameWidth, CV_8UC3);
		///////////////////////////////////////////////////////
		Mat frame;
		if (useCamera)
		{
			Mat frameOrig;
			cap >> frameOrig;
			resize(frameOrig, frame, Size(conf.frameWidth, conf.frameHeight));
			flip(frame, frame, 1);
			frame.copyTo(result);
			if (doInitialise)
			{
				if (tracker.IsInitialised())
				{
					tracker.Reset();
				}
				else
				{
					tracker.InitialiseGPR(frame, initBB, frameInd);
				}
				doInitialise = false;
			}
			else if (!tracker.IsInitialised())
			{
				rectangle(result, initBB, CV_RGB(255, 255, 255));
			}
		}
		else
		{
			char imgPath[256];
			sprintf(imgPath, imgFormat.c_str(), frameInd);
			Mat frameOrig = cv::imread(imgPath, 0);
			if (frameOrig.empty())
			{
				cout << "error: could not read frame: " << imgPath << endl;
				return EXIT_FAILURE;
			}
			resize(frameOrig, frame, Size(conf.frameWidth, conf.frameHeight));
			cvtColor(frame, result, CV_GRAY2RGB);

			if (frameInd == startFrame)
			{
				tracker.InitialiseGPR(frame, initBB, frameInd);  //Initialise TGPR using the first frame, including collection of Target, Auxiliary samples
			}
		}

		if (tracker.IsInitialised())
		{
			ftime(&Timer.startTimer8);
			tracker.Track(frame, frameInd); //Tracking from the first frame
			ftime(&Timer.endTimer8);
			Timer.storetime(Timer.TimeStore8, Timer.startTimer8, Timer.endTimer8);

			if (!conf.quietMode && conf.debugMode)
			{
				tracker.Debug(); //Intermediate tracking results. sampler fig: Unlabeled samples ranked according to the scores of Auxiliary Classifier;
				//sampler1 fig: Unlabeled samples ranked according to the scores of Target Classifier;
				//sampler2 fig: Show Auxiliary Positive, Auxiliary Negative,Target Positive, Target Negative, Unlabeled Samples. Auxiliary samples are shown from flag==1;
				//flag: indicates whether the auxiliary samples achieve a proper size
			}

			rectangle(result, tracker.GetBB(), CV_RGB(0, 255, 0));

			///////////////////////////////////////////////////////
			//SaveResults_1
			//Temp.push_back(result);
			///////////////////////////////////////////////////////

			///////////////////////////////////////////////////////
			//SaveResults_2
			Temp.push_back(tracker.GetBB());
			///////////////////////////////////////////////////////

			cout << "Frame " << frameInd << endl;

			if (outFile)
			{
				const FloatRect& bb = tracker.GetBB();
				outFile << bb.XMin() / scaleW << "," << bb.YMin() / scaleH << "," << bb.Width() / scaleW << "," << bb.Height() / scaleH << endl; //storing the tracking results
			}
		}

		if (!conf.quietMode)
		{
			imshow("result", result); //qualitatively show the tracking results 
			int key = waitKey(paused ? 0 : 1);
			if (key != -1)
			{
				if (key == 27 || key == 113) // esc q
				{
					break;
				}
				else if (key == 112) // p
				{
					paused = !paused;
				}
				else if (key == 105 && useCamera)
				{
					doInitialise = true;
				}
			}
			if (frameInd == endFrame)
			{
				ftime(&endTime);

				/*				cout << "accurate to milliseconds, tracking this sequence costs " << (endTime.time - startTime.time) * 1000 + (endTime.millitm - startTime.millitm) << "ms" << endl;
				if (Timer.TimeStore1.size() != 0)
				{
				double Timer1Total = 0.0;
				for (int i = 1; i < Timer.TimeStore1.size(); ++i)
				{
				Timer1Total += Timer.TimeStore1.at(i);
				}
				cout << "Timer1 " << Timer1Total / Timer.TimeStore1.size() << "s" << endl;
				}
				if (Timer.TimeStore2.size() != 0)
				{
				double Timer2Total = 0.0;
				for (int i = 1; i < Timer.TimeStore2.size(); ++i)
				{
				Timer2Total += Timer.TimeStore2.at(i);
				}
				cout << "Timer2 " << Timer2Total / Timer.TimeStore2.size() << "s" << endl;
				}
				if (Timer.TimeStore3.size() != 0)
				{
				double Timer3Total = 0.0;
				for (int i = 1; i < Timer.TimeStore3.size(); ++i)
				{
				Timer3Total += Timer.TimeStore3.at(i);
				}
				cout << "Timer3: GPR::costruKernel() " << Timer3Total / Timer.TimeStore3.size() << "s" << endl;
				}
				if (Timer.TimeStore4.size() != 0)
				{
				double Timer4Total = 0.0;
				for (int i = 1; i < Timer.TimeStore4.size(); ++i)
				{
				Timer4Total += Timer.TimeStore4.at(i);
				}
				cout << "Timer4 " << Timer4Total / Timer.TimeStore4.size() << "s" << endl;
				}
				if (Timer.TimeStore5.size() != 0)
				{
				double Timer5Total = 0.0;
				for (int i = 1; i < Timer.TimeStore5.size(); ++i)
				{
				Timer5Total += Timer.TimeStore5.at(i);
				}
				cout << "Timer5: the regression process " << Timer5Total / Timer.TimeStore5.size() << "s" << endl;
				}
				if (Timer.TimeStore6.size() != 0)
				{
				double Timer6Total = 0.0;
				for (int i = 1; i < Timer.TimeStore6.size(); ++i)
				{
				Timer6Total += Timer.TimeStore6.at(i);
				}
				cout << "Timer6: log feature extraction of unlabeled samples " << Timer6Total / Timer.TimeStore6.size() << "s" << endl;
				}
				if (Timer.TimeStore7.size() != 0)
				{
				double Timer7Total = 0.0;
				for (int i = 1; i < Timer.TimeStore7.size(); ++i)
				{
				Timer7Total += Timer.TimeStore7.at(i);
				}
				cout << "Timer7: log feature extraction of labeled samples " << Timer7Total / Timer.TimeStore7.size() << "s" << endl;
				}*/
				if (Timer.TimeStore5.size() != 0)
				{
					double Timer5Total = 0.0;
					for (int i = 1; i < Timer.TimeStore5.size(); ++i)
					{
						Timer5Total += Timer.TimeStore5.at(i);
					}
					cout << "Timer5: the regression process " << Timer5Total / Timer.TimeStore5.size() << "s" << endl;
				}
				if (Timer.TimeStore8.size() != 0)
				{
					double Timer8Total = 0.0;
					for (int i = 1; i < Timer.TimeStore8.size(); ++i)
					{
						Timer8Total += Timer.TimeStore8.at(i);
					}
					cout << "time cost of each frame " << Timer8Total / Timer.TimeStore8.size() << "s" << endl;
				}
				cout << "generating the tracking results figures... " << endl;

				string imgFormatwrite = conf.resultsPath + "img%05d.png";
				char imgPathwrite[256];

				///////////////////////////////////////////////////////
				//SaveResults_1
				//for (int i = 0; i < Temp.size(); ++i)
				//{
				//	sprintf(imgPathwrite, imgFormatwrite.c_str(), i);
				//imshow("result", Temp[i]);
				//	cv::imwrite(imgPathwrite, Temp[i]); //0 indicates gray chanal
				//}
				///////////////////////////////////////////////////////

				///////////////////////////////////////////////////////
				//SaveResults_2
				Mat resultwrite(conf.frameHeight, conf.frameWidth, CV_8UC3);
				for (int i = 0; i < Temp.size(); ++i)
				{
					Mat framewrite;
					char imgPathwrite[256];
					sprintf(imgPathwrite, imgFormat.c_str(), startFrame + i);
					Mat frameOrigwrite = cv::imread(imgPathwrite, 0);
					if (frameOrigwrite.empty())
					{
						cout << "error: could not read frame: " << imgPathwrite << endl;
						return EXIT_FAILURE;
					}
					resize(frameOrigwrite, framewrite, Size(conf.frameWidth, conf.frameHeight));
					cvtColor(framewrite, resultwrite, CV_GRAY2RGB);

					rectangle(resultwrite, Temp[i], CV_RGB(0, 255, 0));
					sprintf(imgPathwrite, imgFormatwrite.c_str(), startFrame + i);
					//imshow("result", Temp[i]);
					cv::imwrite(imgPathwrite, resultwrite); //0 indicates gray chanal
				}
				///////////////////////////////////////////////////////

				//cout << "\n\nend of sequence, press esc or q to exit" << endl;
				//waitKey();
			}
			if (conf.debugMode && frameInd == endFrame)
			{
				cout << "\n\nend of sequence, press any key to exit" << endl;
				waitKey();
			}
		}
	}


	if (outFile.is_open())
	{
		outFile.close();
	}

	return EXIT_SUCCESS;
}
